import datetime
import logging
import sqlite3
from collections import namedtuple
from functools import lru_cache
from pathlib import Path
from typing import Self, Any
import polars as pl
import numpy as np
import pandas as pd
import structlog
from numpy import integer as any_integer
from sqlalchemy.ext.asyncio import AsyncSession, async_sessionmaker, create_async_engine

from ziplime.assets.entities.asset import Asset
from ziplime.utils.functional import keysorted
from ziplime.utils.numpy_utils import (
    datetime64ns_dtype,
    float64_dtype,
    int64_dtype,
    uint32_dtype,
    uint64_dtype,
)
from ziplime.utils.pandas_utils import empty_dataframe
from ziplime.utils.sqlite_utils import group_into_chunks

from ziplime.data.adjustments import load_adjustments_from_sqlite

from ziplime.assets.repositories.adjustments_repository import AdjustmentRepository

log = structlog.get_logger(__name__)

SQLITE_ADJUSTMENT_TABLENAMES = frozenset(["splits", "dividends", "mergers"])

UNPAID_QUERY_TEMPLATE = """
                        SELECT sid, amount, pay_date
                        from dividend_payouts
                        WHERE ex_date = ?
                          AND sid IN ({0}) \
                        """

Dividend = namedtuple("Dividend", ["asset", "amount", "pay_date"])

UNPAID_STOCK_DIVIDEND_QUERY_TEMPLATE = """
                                       SELECT sid, payment_sid, ratio, pay_date
                                       from stock_dividend_payouts
                                       WHERE ex_date = ?
                                         AND sid IN ({0}) \
                                       """

StockDividend = namedtuple(
    "StockDividend",
    ["asset", "payment_asset", "ratio", "pay_date"],
)

SQLITE_ADJUSTMENT_COLUMN_DTYPES = {
    "effective_date": any_integer,
    "ratio": float64_dtype,
    "sid": any_integer,
}

SQLITE_DIVIDEND_PAYOUT_COLUMN_DTYPES = {
    "sid": any_integer,
    "ex_date": any_integer,
    "declared_date": any_integer,
    "record_date": any_integer,
    "pay_date": any_integer,
    "amount": float,
}

SQLITE_STOCK_DIVIDEND_PAYOUT_COLUMN_DTYPES = {
    "sid": any_integer,
    "ex_date": any_integer,
    "declared_date": any_integer,
    "record_date": any_integer,
    "pay_date": any_integer,
    "payment_sid": any_integer,
    "ratio": float,
}


def specialize_any_integer(d):
    out = {}
    for k, v in d.items():
        if v is any_integer:
            out[k] = int64_dtype
        else:
            out[k] = v
    return out


class SqlAlchemyAdjustmentRepository(AdjustmentRepository):
    """Loads adjustments based on corporate actions from a SQLite database.

    Expects data written in the format output by `SQLiteAdjustmentWriter`.

    Parameters
    ----------
    conn : str or sqlite3.Connection
        Connection from which to load data.

    See Also
    --------
    :class:`ziplime.data.adjustments.SQLiteAdjustmentWriter`
    """

    _datetime_int_cols = {
        "splits": ("effective_date",),
        "mergers": ("effective_date",),
        "dividends": ("effective_date",),
        "dividend_payouts": (
            "declared_date",
            "ex_date",
            "pay_date",
            "record_date",
        ),
        "stock_dividend_payouts": (
            "declared_date",
            "ex_date",
            "pay_date",
            "record_date",
        ),
    }
    _raw_table_dtypes = {
        # We use any_integer above to be lenient in accepting different dtypes
        # from users. For our outputs, however, we always want to return the
        # same types, and any_integer turns into int32 on some numpy windows
        # builds, so specify int64 explicitly here.
        "splits": specialize_any_integer(SQLITE_ADJUSTMENT_COLUMN_DTYPES),
        "mergers": specialize_any_integer(SQLITE_ADJUSTMENT_COLUMN_DTYPES),
        "dividends": specialize_any_integer(SQLITE_ADJUSTMENT_COLUMN_DTYPES),
        "dividend_payouts": specialize_any_integer(
            SQLITE_DIVIDEND_PAYOUT_COLUMN_DTYPES,
        ),
        "stock_dividend_payouts": specialize_any_integer(
            SQLITE_STOCK_DIVIDEND_PAYOUT_COLUMN_DTYPES,
        ),
    }

    def __init__(self, db_url: str):
        self.db_url = db_url

    def __enter__(self):
        return self

    @property
    @lru_cache
    def session_maker(self) -> async_sessionmaker[AsyncSession]:
        engine = create_async_engine(self.db_url, pool_pre_ping=True, pool_size=20)
        session_maker = async_sessionmaker(autocommit=False, autoflush=True, bind=engine, class_=AsyncSession,
                                           expire_on_commit=False)
        return session_maker

    def load_adjustments(
            self,
            dates,
            assets,
            should_include_splits,
            should_include_mergers,
            should_include_dividends,
            adjustment_type,
    ):
        """Load collection of Adjustment objects from underlying adjustments db.

        Parameters
        ----------
        dates : pd.DatetimeIndex
            Dates for which adjustments are needed.
        assets : pd.Int64Index
            Assets for which adjustments are needed.
        should_include_splits : bool
            Whether split adjustments should be included.
        should_include_mergers : bool
            Whether merger adjustments should be included.
        should_include_dividends : bool
            Whether dividend adjustments should be included.
        adjustment_type : str
            Whether price adjustments, volume adjustments, or both, should be
            included in the output.

        Returns
        -------
        adjustments : dict[str -> dict[int -> Adjustment]]
            A dictionary containing price and/or volume adjustment mappings
            from index to adjustment objects to apply at that index.
        """
        dates = dates.tz_localize("UTC")
        return load_adjustments_from_sqlite(
            self.conn,
            dates,
            assets,
            should_include_splits,
            should_include_mergers,
            should_include_dividends,
            adjustment_type,
        )

    def load_pricing_adjustments(self, columns, dates, assets):
        if "volume" not in set(columns):
            adjustment_type = "price"
        elif len(set(columns)) == 1:
            adjustment_type = "volume"
        else:
            adjustment_type = "all"

        adjustments = self.load_adjustments(
            dates,
            assets,
            should_include_splits=True,
            should_include_mergers=True,
            should_include_dividends=True,
            adjustment_type=adjustment_type,
        )
        price_adjustments = adjustments.get("price")
        volume_adjustments = adjustments.get("volume")

        return [
            volume_adjustments if column == "volume" else price_adjustments
            for column in columns
        ]

    def get_adjustments_for_sid(self, table_name, sid):
        return []
        t = (sid,)
        c = self.conn.cursor()
        adjustments_for_sid = c.execute(
            "SELECT effective_date, ratio FROM %s WHERE sid = ?" % table_name, t
        ).fetchall()
        c.close()

        return [
            [pd.Timestamp(adjustment[0], unit="s"), adjustment[1]]
            for adjustment in adjustments_for_sid
        ]

    def get_dividends_with_ex_date(self, assets, date, asset_finder):
        # seconds = date.value / int(1e9)
        return []
        c = self.conn.cursor()

        divs = []
        for chunk in group_into_chunks(assets):
            query = UNPAID_QUERY_TEMPLATE.format(",".join(["?" for _ in chunk]))
            t = (date,) + tuple(map(lambda x: int(x), chunk))

            c.execute(query, t)

            rows = c.fetchall()
            for row in rows:
                div = Dividend(
                    asset_finder.retrieve_asset(row[0]),
                    row[1],
                    pd.Timestamp(row[2], unit="s", tz="UTC"),
                )
                divs.append(div)
        c.close()

        return divs

    def get_stock_dividends(self, sid: int, trading_days: pl.Series) -> list[Dividend]:
        return []

    def get_stock_dividends_with_ex_date(self, assets, date, asset_finder):
        # seconds = date.value / int(1e9)
        return []

        c = self.conn.cursor()

        stock_divs = []
        for chunk in group_into_chunks(assets):
            query = UNPAID_STOCK_DIVIDEND_QUERY_TEMPLATE.format(
                ",".join(["?" for _ in chunk])
            )
            t = (date,) + tuple(map(lambda x: int(x), chunk))

            c.execute(query, t)

            rows = c.fetchall()

            for row in rows:
                stock_div = StockDividend(
                    asset_finder.retrieve_asset(row[0]),  # asset
                    asset_finder.retrieve_asset(row[1]),  # payment_asset
                    row[2],
                    pd.Timestamp(row[3], unit="s", tz="UTC"),
                )
                stock_divs.append(stock_div)
        c.close()

        return stock_divs

    def unpack_db_to_component_dfs(self, convert_dates=False):
        """Returns the set of known tables in the adjustments file in DataFrame
        form.

        Parameters
        ----------
        convert_dates : bool, optional
            By default, dates are returned in seconds since EPOCH. If
            convert_dates is True, all ints in date columns will be converted
            to datetimes.

        Returns
        -------
        dfs : dict{str->DataFrame}
            Dictionary which maps table name to the corresponding DataFrame
            version of the table, where all date columns have been coerced back
            from int to datetime.
        """
        return {
            t_name: self.get_df_from_table(t_name, convert_dates)
            for t_name in self._datetime_int_cols
        }

    def get_df_from_table(self, table_name, convert_dates=False):
        try:
            date_cols = self._datetime_int_cols[table_name]
        except KeyError as exc:
            raise ValueError(
                f"Requested table {table_name} not found.\n"
                f"Available tables: {self._datetime_int_cols.keys()}\n"
            ) from exc

        # Dates are stored in second resolution as ints in adj.db tables.
        kwargs = (
            # {"parse_dates": {col: {"unit": "s", "utc": True} for col in date_cols}}
            {"parse_dates": {col: {"unit": "s"} for col in date_cols}}
            if convert_dates
            else {}
        )

        result = pd.read_sql(
            f"select * from {table_name}",
            self.conn,
            index_col="index",
            **kwargs,
        )
        dtypes = self._df_dtypes(table_name, convert_dates)

        if not len(result):
            return empty_dataframe(*keysorted(dtypes))

        result.rename_axis(None, inplace=True)
        result = result[sorted(dtypes)]  # ensure expected order of columns
        return result

    def _df_dtypes(self, table_name, convert_dates):
        """Get dtypes to use when unpacking sqlite tables as dataframes."""
        out = self._raw_table_dtypes[table_name]
        if convert_dates:
            out = out.copy()
            for date_column in self._datetime_int_cols[table_name]:
                out[date_column] = datetime64ns_dtype

        return out

    """Writer for data to be read by SQLiteAdjustmentReader

    Parameters
    ----------
    conn_or_path : str or sqlite3.Connection
        A handle to the target sqlite database.
    equity_daily_bar_reader : SessionBarReader
        Daily bar reader to use for dividend writes.
    overwrite : bool, optional, default=False
        If True and conn_or_path is a string, remove any existing files at the
        given path before connecting.

    See Also
    --------
    ziplime.data.adjustments.SQLiteAdjustmentReader
    """

    def _write(self, tablename, expected_dtypes, frame):
        if frame is None or frame.empty:
            # keeping the dtypes correct for empty frames is not easy
            # frame = pd.DataFrame(
            #     np.array([], dtype=list(expected_dtypes.items())),
            # )
            frame = pd.DataFrame(expected_dtypes, index=[])
        else:
            if frozenset(frame.columns) != frozenset(expected_dtypes):
                raise ValueError(
                    "Unexpected frame columns:\n"
                    "Expected Columns: %s\n"
                    "Received Columns: %s"
                    % (
                        set(expected_dtypes),
                        frame.columns.tolist(),
                    )
                )

            actual_dtypes = frame.dtypes
            for colname, expected in expected_dtypes.items():
                actual = actual_dtypes[colname]
                if not np.issubdtype(actual, expected):
                    raise TypeError(
                        "Expected data of type {expected} for column"
                        " '{colname}', but got '{actual}'.".format(
                            expected=expected,
                            colname=colname,
                            actual=actual,
                        ),
                    )

        frame.to_sql(
            tablename,
            self.conn,
            if_exists="append",
            chunksize=50000,
        )

    def write_frame(self, tablename, frame):
        if tablename not in SQLITE_ADJUSTMENT_TABLENAMES:
            raise ValueError(
                f"Adjustment table {tablename} not in {SQLITE_ADJUSTMENT_TABLENAMES}"
            )
        if not (frame is None or frame.empty):
            frame = frame.copy()
            frame["effective_date"] = (
                frame["effective_date"]
                .values.astype(
                    "datetime64[s]",
                )
                .astype("int64")
            )
        return self._write(
            tablename,
            SQLITE_ADJUSTMENT_COLUMN_DTYPES,
            frame,
        )

    def write_dividend_payouts(self, frame):
        """Write dividend payout data to SQLite table `dividend_payouts`."""
        return self._write(
            "dividend_payouts",
            SQLITE_DIVIDEND_PAYOUT_COLUMN_DTYPES,
            frame,
        )

    def write_stock_dividend_payouts(self, frame):
        return self._write(
            "stock_dividend_payouts",
            SQLITE_STOCK_DIVIDEND_PAYOUT_COLUMN_DTYPES,
            frame,
        )

    def calc_dividend_ratios(self, dividends):
        """Calculate the ratios to apply to equities when looking back at pricing
        history so that the price is smoothed over the ex_date, when the market
        adjusts to the change in equity value due to upcoming dividend.

        Returns
        -------
        DataFrame
            A frame in the same format as splits and mergers, with keys
            - sid, the id of the equity
            - effective_date, the date in seconds on which to apply the ratio.
            - ratio, the ratio to apply to backwards looking pricing data.
        """
        if dividends is None or dividends.empty:
            return pd.DataFrame(
                np.array(
                    [],
                    dtype=[
                        ("sid", uint64_dtype),
                        ("effective_date", uint32_dtype),
                        ("ratio", float64_dtype),
                    ],
                )
            )

        pricing_reader = self._equity_daily_bar_reader
        input_sids = dividends.sid.values
        unique_sids, sids_ix = np.unique(input_sids, return_inverse=True)
        dates = pricing_reader.sessions.values

        (close,) = pricing_reader.load_raw_arrays(
            ["close"],
            pd.Timestamp(dates[0]),
            pd.Timestamp(dates[-1]),
            unique_sids,
        )
        date_ix = np.searchsorted(dates, dividends.ex_date.values)
        mask = date_ix > 0

        date_ix = date_ix[mask]
        sids_ix = sids_ix[mask]
        input_dates = dividends.ex_date.values[mask]

        # subtract one day to get the close on the day prior to the merger
        previous_close = close[date_ix - 1, sids_ix]
        input_sids = input_sids[mask]

        amount = dividends.amount.values[mask]
        ratio = 1.0 - amount / previous_close

        non_nan_ratio_mask = ~np.isnan(ratio)
        for ix in np.flatnonzero(~non_nan_ratio_mask):
            log.warning(
                "Couldn't compute ratio for dividend"
                " sid=%(sid)s, ex_date=%(ex_date)s, amount=%(amount).3f",
                {
                    "sid": input_sids[ix],
                    "ex_date": pd.Timestamp(input_dates[ix]).strftime("%Y-%m-%d"),
                    "amount": amount[ix],
                },
            )

        positive_ratio_mask = ratio > 0
        for ix in np.flatnonzero(~positive_ratio_mask & non_nan_ratio_mask):
            log.warning(
                "Dividend ratio <= 0 for dividend"
                " sid=%(sid)s, ex_date=%(ex_date)s, amount=%(amount).3f",
                {
                    "sid": input_sids[ix],
                    "ex_date": pd.Timestamp(input_dates[ix]).strftime("%Y-%m-%d"),
                    "amount": amount[ix],
                },
            )

        valid_ratio_mask = non_nan_ratio_mask & positive_ratio_mask
        return pd.DataFrame(
            {
                "sid": input_sids[valid_ratio_mask],
                "effective_date": input_dates[valid_ratio_mask],
                "ratio": ratio[valid_ratio_mask],
            }
        )

    def _write_dividends(self, dividends):
        if dividends is None:
            dividend_payouts = None
        else:
            dividend_payouts = dividends.copy()
            # TODO: Check if that's the right place for this fix for pandas > 1.2.5
            dividend_payouts.fillna(np.datetime64("NaT"), inplace=True)
            dividend_payouts["ex_date"] = (
                dividend_payouts["ex_date"]
                .values.astype("datetime64[s]")
                .astype(int64_dtype)
            )
            dividend_payouts["record_date"] = (
                dividend_payouts["record_date"]
                .values.astype("datetime64[s]")
                .astype(int64_dtype)
            )
            dividend_payouts["declared_date"] = (
                dividend_payouts["declared_date"]
                .values.astype("datetime64[s]")
                .astype(int64_dtype)
            )
            dividend_payouts["pay_date"] = (
                dividend_payouts["pay_date"]
                .values.astype("datetime64[s]")
                .astype(int64_dtype)
            )

        self.write_dividend_payouts(dividend_payouts)

    def _write_stock_dividends(self, stock_dividends):
        if stock_dividends is None:
            stock_dividend_payouts = None
        else:
            stock_dividend_payouts = stock_dividends.copy()
            stock_dividend_payouts["ex_date"] = (
                stock_dividend_payouts["ex_date"]
                .values.astype("datetime64[s]")
                .astype(int64_dtype)
            )
            stock_dividend_payouts["record_date"] = (
                stock_dividend_payouts["record_date"]
                .values.astype("datetime64[s]")
                .astype(int64_dtype)
            )
            stock_dividend_payouts["declared_date"] = (
                stock_dividend_payouts["declared_date"]
                .values.astype("datetime64[s]")
                .astype(int64_dtype)
            )
            stock_dividend_payouts["pay_date"] = (
                stock_dividend_payouts["pay_date"]
                .values.astype("datetime64[s]")
                .astype(int64_dtype)
            )
        self.write_stock_dividend_payouts(stock_dividend_payouts)

    def write_dividend_data(self, dividends, stock_dividends=None):
        """Write both dividend payouts and the derived price adjustment ratios."""

        # First write the dividend payouts.
        self._write_dividends(dividends)
        self._write_stock_dividends(stock_dividends)

        # Second from the dividend payouts, calculate ratios.
        dividend_ratios = self.calc_dividend_ratios(dividends)
        self.write_frame("dividends", dividend_ratios)

    def write(self, splits=None, mergers=None, dividends=None, stock_dividends=None):
        """Writes data to a SQLite file to be read by SQLiteAdjustmentReader.

        Parameters
        ----------
        splits : pandas.DataFrame, optional
            Dataframe containing split data. The format of this dataframe is:
              effective_date : int
                  The date, represented as seconds since Unix epoch, on which
                  the adjustment should be applied.
              ratio : float
                  A value to apply to all data earlier than the effective date.
                  For open, high, low, and close those values are multiplied by
                  the ratio. Volume is divided by this value.
              sid : int
                  The asset id associated with this adjustment.
        mergers : pandas.DataFrame, optional
            DataFrame containing merger data. The format of this dataframe is:
              effective_date : int
                  The date, represented as seconds since Unix epoch, on which
                  the adjustment should be applied.
              ratio : float
                  A value to apply to all data earlier than the effective date.
                  For open, high, low, and close those values are multiplied by
                  the ratio. Volume is unaffected.
              sid : int
                  The asset id associated with this adjustment.
        dividends : pandas.DataFrame, optional
            DataFrame containing dividend data. The format of the dataframe is:
              sid : int
                  The asset id associated with this adjustment.
              ex_date : datetime64
                  The date on which an equity must be held to be eligible to
                  receive payment.
              declared_date : datetime64
                  The date on which the dividend is announced to the public.
              pay_date : datetime64
                  The date on which the dividend is distributed.
              record_date : datetime64
                  The date on which the stock ownership is checked to determine
                  distribution of dividends.
              amount : float
                  The cash amount paid for each share.

            Dividend ratios are calculated as:
            ``1.0 - (dividend_value / "close on day prior to ex_date")``
        stock_dividends : pandas.DataFrame, optional
            DataFrame containing stock dividend data. The format of the
            dataframe is:
              sid : int
                  The asset id associated with this adjustment.
              ex_date : datetime64
                  The date on which an equity must be held to be eligible to
                  receive payment.
              declared_date : datetime64
                  The date on which the dividend is announced to the public.
              pay_date : datetime64
                  The date on which the dividend is distributed.
              record_date : datetime64
                  The date on which the stock ownership is checked to determine
                  distribution of dividends.
              payment_sid : int
                  The asset id of the shares that should be paid instead of
                  cash.
              ratio : float
                  The ratio of currently held shares in the held sid that
                  should be paid with new shares of the payment_sid.

        See Also
        --------
        ziplime.data.adjustments.SQLiteAdjustmentReader
        """
        self.write_frame("splits", splits)
        self.write_frame("mergers", mergers)
        self.write_dividend_data(dividends, stock_dividends)

    def get_splits(self, assets: frozenset[Asset], dt: datetime.date):
        """Returns any splits for the given sids and the given dt.

        Parameters
        ----------
        assets : container
            Assets for which we want splits.
        dt : datetime.datetime
            The date for which we are checking for splits. Note: this is
            expected to be midnight UTC.

        Returns
        -------
        splits : list[(asset, float)]
            List of splits, where each split is a (asset, ratio) tuple.
        """
        return []
        if not assets:
            return []

        # convert dt to # of seconds since epoch, because that's what we use
        # in the adjustments db
        # seconds = int(dt.value / 1e9)

        splits = self.adjustment_repository.conn.execute(
            "SELECT sid, ratio FROM SPLITS WHERE effective_date = ?", (dt,)
        ).fetchall()

        splits = [split for split in splits if split[0] in assets]
        splits = [
            (self.asset_repository.retrieve_asset(split[0]), split[1]) for split in splits
        ]

        return splits

    def to_json(self):
        return {
            "base_storage_path": self._base_storage_path,
            "bundle_name": self._bundle_name,
            "bundle_version": self._bundle_version,
        }

    @classmethod
    def from_json(cls, data: dict[str, Any]) -> Self:
        return cls(
            base_storage_path=data["base_storage_path"],
            bundle_name=data["bundle_name"],
            bundle_version=data["bundle_version"],
        )



import datetime
import multiprocessing
import os
import sys
import yfinance as yf
from typing import Self

import limexhub
import structlog
from asyncclick import progressbar
from joblib import Parallel, delayed

import polars as pl
from ziplime.data.services.data_bundle_source import DataBundleSource


class YahooFinanceDataSource(DataBundleSource):
    def __init__(self, maximum_threads: int | None = None):
        super().__init__()
        self._logger = structlog.get_logger(__name__)
        if maximum_threads is not None:
            self._maximum_threads = min(multiprocessing.cpu_count() * 2, maximum_threads)
        else:
            self._maximum_threads = multiprocessing.cpu_count() * 2

    def _get_frequency(self, frequency: datetime.timedelta) -> str:
        if frequency == datetime.timedelta(minutes=1):
            return "1m"
        elif frequency == datetime.timedelta(minutes=2):
            return "2m"
        elif frequency == datetime.timedelta(minutes=5):
            return "5m"
        elif frequency == datetime.timedelta(minutes=15):
            return "15m"
        elif frequency == datetime.timedelta(minutes=30):
            return "30m"
        elif frequency == datetime.timedelta(minutes=60):
            return "60m"
        elif frequency == datetime.timedelta(minutes=90):
            return "90m"
        elif frequency == datetime.timedelta(days=1):
            return "1d"
        elif frequency == datetime.timedelta(days=5):
            return "5d"

        raise ValueError(f"Unsupported frequency for Yahoo Finance {frequency}")

    async def get_data(self, symbols: list[str],
                       frequency: datetime.timedelta,
                       date_from: datetime.datetime,
                       date_to: datetime.datetime,
                       **kwargs
                       ) -> pl.DataFrame:

        yfinance_data_raw = yf.download(tickers=symbols, threads=1,
                                        interval=self._get_frequency(frequency=frequency),
                                        start=date_from, end=date_to, group_by="Ticker", progress=True,
                                        auto_adjust=True,
                                        multi_level_index=False)
        final = pl.DataFrame()
        for symbol in symbols:
            df_symbol = yfinance_data_raw[symbol]

            df = pl.from_pandas(df_symbol, include_index=True,
                                schema_overrides={"Open": pl.Float64(), "High": pl.Float64(),
                                                  "Low": pl.Float64(), "Close": pl.Float64(),
                                                  "Volume": pl.Float64()}
                                )
            if "Date" in df.columns:
                date_column = "Date"
            else:
                date_column = "Datetime"

            if len(df) > 0:
                df = df.rename(
                    {
                        "Open": "open",
                        "High": "high",
                        "Low": "low",
                        "Close": "close",
                        "Volume": "volume",
                        f"{date_column}": "date"
                    }
                )
                df = df.with_columns(
                    pl.lit(symbol).alias("symbol"),
                    pl.lit("LIME").alias("exchange"),
                    pl.lit("US").alias("exchange_country"),
                    pl.col("close").alias("price"),
                )
                if date_column == "Datetime":
                    df = df.with_columns(date=pl.col("date").dt.convert_time_zone(str(date_from.tzinfo)))
                else:
                    df = df.with_columns(date=pl.col("date").dt.replace_time_zone(str(date_from.tzinfo)))
                df = df.filter(pl.col("date") >= date_from, pl.col("date") <= date_to)
                final = pl.concat([final, df])
        return final



from ziplime.finance.shared import AllowedAssetMarker

from ziplime.assets.entities.equity import Equity
from ziplime.finance.slippage.slippage_model import SlippageModel


class EquitySlippageModel(SlippageModel, metaclass=AllowedAssetMarker):
    """Base class for slippage models which only support equities."""

    allowed_asset_types = (Equity,)



from sqlalchemy.orm import Mapped

from ziplime.core.db.annotated_types import StringPK
from ziplime.core.db.base_model import BaseModel


class ExchangeInfo(BaseModel):
    __tablename__ = "exchanges"

    exchange: Mapped[StringPK]
    canonical_name: Mapped[str]
    country_code: Mapped[str]



from typing import Any

import pandas as pd

from ziplime.trading.trading_algorithm import TradingAlgorithm
from ziplime.domain.bar_data import BarData


def get_fundamental_data(
        bar_data: BarData,
        context: TradingAlgorithm,
        assets,
        fields,
        bar_count,
        frequency,
        fillna: Any = None
):
    if frequency != '1q':
        raise Exception("Currently only frequency of 1 quarter is supported for fundamental data")

    max_days_per_quarter = 92
    end_date = pd.to_datetime(bar_data.current_session)
    start_date = end_date - pd.Timedelta(days=max_days_per_quarter * bar_count)

    fundamental_data_bundle = context.fundamental_data_bundle
    first_date = fundamental_data_bundle.first_trading_day

    start_date_array = start_date
    if start_date < first_date:
        start_date_array = first_date

    dr = pd.date_range(start_date_array, bar_data.current_session, freq='D')

    res = fundamental_data_bundle.load_raw_arrays_full_range(
        [fields],
        start_date_array,
        bar_data.current_session,
        [assets],
        frequency='D'
    )[0]

    df = pd.DataFrame(data=res, index=dr)
    quarters = pd.date_range(start=start_date, end=end_date, freq='QE')
    df = df.reindex(quarters)
    if fillna is not None:
        df[df.isnull()] = fillna

    result_series = pd.Series(data=df[df.columns[0]], index=df.index)
    return result_series



class ContractNode:

    def __init__(self, contract):
        self.contract = contract
        self.prev = None
        self.next = None

    def __rshift__(self, offset):
        i = 0
        curr = self
        while i < offset and curr is not None:
            curr = curr.next
            i += 1
        return curr

    def __lshift__(self, offset):
        i = 0
        curr = self
        while i < offset and curr is not None:
            curr = curr.prev
            i += 1
        return curr



import datetime
import polars as pl
import structlog
from exchange_calendars import ExchangeCalendar

from ziplime.assets.services.asset_service import AssetService
from ziplime.exchanges.exchange import Exchange
from ziplime.finance.domain.ledger import Ledger
from ziplime.domain.data_frequency import DataFrequency
from ziplime.sources.benchmark_source import BenchmarkSource


class MetricsTracker:
    """The algorithm's interface to the registered risk and performance
    metrics.

    Parameters
    ----------
    trading_calendar : TradingCalendar
        The trading calendar used in the simulation.
    first_session : datetime.datetime
        The label of the first trading session in the simulation.
    last_session : datetime.datetime
        The label of the last trading session in the simulation.
    capital_base : float
        The starting capital for the simulation.
    emission_rate : {'daily', 'minute'}
        How frequently should a performance packet be generated?
    data_frequency : {'daily', 'minute'}
        The data frequency of the data portal.
    metrics : list[Metric]
        The metrics to track.
    """

    def __init__(
            self,
            sessions: pl.Series,
            asset_service: AssetService,
            exchanges: dict[str, Exchange],
            trading_calendar: ExchangeCalendar,
            first_session: datetime.datetime,
            last_session: datetime.datetime,
            emission_rate: datetime.timedelta,
            ledger: Ledger,
            metrics,
            benchmark_source: BenchmarkSource,
    ):
        self.emission_rate = emission_rate
        self._benchmark_source = benchmark_source
        self._logger = structlog.getLogger(__name__)

        self._trading_calendar = trading_calendar
        self._first_session = first_session
        self._last_session = last_session

        self._current_session = first_session
        self._market_open = trading_calendar.session_first_minute(first_session)
        self.exchanges = exchanges
        self.asset_service = asset_service
        # self.data_bundle = data_bundle
        self._session_count = 0

        self._sessions = sessions
        self._total_session_count = len(sessions)
        self._ledger = ledger
        self._metrics = metrics

        self._start_of_simulation_metrics = [
            metric for metric in self._metrics if getattr(metric, "start_of_simulation", None)
        ]
        self._end_of_simulation_metrics = [
            metric for metric in self._metrics if getattr(metric, "end_of_simulation", None)
        ]
        self._start_of_session_metrics = [
            metric for metric in self._metrics if getattr(metric, "start_of_session", None)
        ]
        self._end_of_session_metrics = [
            metric for metric in self._metrics if getattr(metric, "end_of_session", None)
        ]
        self._end_of_bar_metrics = [
            metric for metric in self._metrics if getattr(metric, "end_of_bar", None)
        ]

        if emission_rate == DataFrequency.MINUTE:

            def progress(self):
                return 1.0  # a fake value

        else:

            def progress(self):
                return self._session_count / self._total_session_count

        # don't compare these strings over and over again!
        self._progress = progress

    def handle_start_of_simulation(self):
        for metric in self._start_of_simulation_metrics:
            metric.start_of_simulation(
                ledger=self._ledger,
                emission_rate=self.emission_rate,
                trading_calendar=self._trading_calendar,
                sessions=self._sessions,
                benchmark_source=self._benchmark_source,
            )

    def handle_minute_close(self, dt: datetime.datetime):
        """Handles the close of the given minute in minute emission.

        Parameters
        ----------
        dt : Timestamp
            The minute that is ending

        Returns
        -------
        A minute perf packet.
        """
        self._ledger.sync_last_sale_prices(dt=dt, handle_non_market_minutes=False)
        packet = {
            "period_start": self._first_session,
            "period_end": self._last_session,
            "capital_base": list(self.exchanges.values())[0].get_start_cash_balance(), # TODO: add support for multiple exchanges
            "minute_perf": {
                "period_open": self._market_open,
                "period_close": dt,
            },
            "cumulative_perf": {
                "period_open": self._first_session,
                "period_close": self._last_session,
            },
            "progress": self._progress(self),
            "cumulative_risk_metrics": {},
        }
        ledger = self._ledger
        ledger.end_of_bar(session_ix=self._session_count)
        for metric in self._end_of_bar_metrics:
            metric.end_of_bar(
                packet=packet,
                ledger=ledger,
                session=dt,
                session_ix=self._session_count,
                exchanges=self.exchanges,
            )
        return packet

    def handle_market_open(self, session_label: datetime.datetime) -> None:
        """Handles the start of each session.

        Parameters
        ----------
        session_label : Timestamp
            The label of the session that is about to begin.
        """
        self._ledger.start_of_session(session_label=session_label)

        # TODO: handle ajustments repository
        adjustment_reader = self.asset_service._adjustments_repository
        if adjustment_reader is not None:
            # this is None when running with a dataframe source
            self._ledger.process_dividends(
                next_session=session_label,
                adjustment_reader=adjustment_reader,
            )

        self._current_session = session_label
        self._market_open = self._trading_calendar.session_first_minute(session_label)

        for metric in self._start_of_session_metrics:
            metric.start_of_session(ledger=self._ledger, session=session_label, exchanges=self.exchanges)
        # self.start_of_session(ledger=self._ledger, session=session_label, data_bundle=data_bundle)

    def handle_market_close(self, dt: datetime.datetime):
        """Handles the close of the given day.

        Parameters
        ----------
        dt : Timestamp
            The most recently completed simulation datetime.
        Returns
        -------
        A daily perf packet.
        """

        if self.emission_rate == datetime.timedelta(days=1):
            # this method is called for both minutely and daily emissions, but
            # this chunk of code here only applies for daily emissions. (since
            # it's done every minute, elsewhere, for minutely emission).
            self._ledger.sync_last_sale_prices(dt=dt, handle_non_market_minutes=False)

        session_ix = self._session_count
        # increment the day counter before we move markers forward.
        self._session_count += 1

        packet = {
            "period_start": self._first_session,
            "period_end": self._last_session,
            "capital_base": list(self.exchanges.values())[0].get_start_cash_balance(), # TODO: add support for multiple exchanges
            "daily_perf": {
                "period_open": self._market_open,
                "period_close": dt,
            },
            "cumulative_perf": {
                "period_open": self._first_session,
                "period_close": self._last_session,
            },
            "progress": self._progress(self),
            "cumulative_risk_metrics": {},
        }
        self._ledger.end_of_session(session_ix=session_ix)

        for metric in self._end_of_session_metrics:
            metric.end_of_session(
                packet=packet,
                ledger=self._ledger,
                session=self._current_session,
                session_ix=session_ix,
                exchanges=self.exchanges,
            )
        return packet

    def handle_simulation_end(self):
        """When the simulation is complete, run the full period risk report
        and send it out on the results socket.
        """
        self._logger.info(
            f"Simulated {self._session_count} trading days\n first open: "
            f"{self._trading_calendar.session_open(self._first_session).astimezone(tz=self._trading_calendar.tz)}\n "
            f"last close: {self._trading_calendar.session_close(self._last_session).astimezone(tz=self._trading_calendar.tz)}",
        )

        packet = {}
        for metric in self._end_of_simulation_metrics:
            metric.end_of_simulation(
                packet=packet,
                ledger=self._ledger,
                trading_calendar=self._trading_calendar,
                sessions=self._sessions,
                data_bundle=self.data_bundle,
                benchmark_source=self._benchmark_source,
            )
        return packet



import datetime

from sqlalchemy.orm import Mapped

from ziplime.assets.models.asset_model import AssetModel
from ziplime.core.db.annotated_types import StringUnique, FuturesRootSymbolFK, ExchangeFK


class FuturesContractModel(AssetModel):
    __tablename__ = "futures_contracts"

    symbol: Mapped[StringUnique]
    root_symbol: Mapped[FuturesRootSymbolFK]
    notice_date: Mapped[datetime.date]
    expiration_date: Mapped[datetime.date]
    multiplier: Mapped[float]
    tick_size: Mapped[float]
    exchange: Mapped[ExchangeFK]



import datetime
from typing import Any

import empyrical as ep
import numpy as np
import pandas as pd
from exchange_calendars import ExchangeCalendar


from ziplime.exchanges.exchange import Exchange
from ziplime.finance.domain.ledger import Ledger
from ziplime.sources.benchmark_source import BenchmarkSource



class AlphaBeta:
    """End of simulation alpha and beta to the benchmark."""

    def start_of_simulation(
            self, ledger: Ledger, emission_rate: datetime.timedelta, trading_calendar: ExchangeCalendar,
            sessions: pd.DatetimeIndex, benchmark_source: BenchmarkSource
    ):
        self._daily_returns_array = benchmark_source.daily_returns(
            sessions[0],
            sessions[-1],
        ).select("pct_change")

    def end_of_bar(self, packet: dict[str, Any], ledger: Ledger, session: datetime.datetime, session_ix: int,
                   exchanges: dict[str, Exchange]):
        risk = packet["cumulative_risk_metrics"]
        alpha, beta = ep.alpha_beta_aligned(
            ledger.daily_returns_array[: session_ix + 1],
            self._daily_returns_array[: session_ix + 1]["pct_change"].to_pandas(),
        )
        if not np.isfinite(alpha):
            alpha = None
        if np.isnan(beta):
            beta = None

        risk["alpha"] = alpha
        risk["beta"] = beta

    end_of_session = end_of_bar



import datetime
from dataclasses import dataclass

from ziplime.trading.entities.trading_pair import TradingPair
from ziplime.trading.enums.order_type import OrderType


@dataclass
class OrderRequest:
    order_id: str
    trading_pair: TradingPair
    order_type: OrderType
    creation_date: datetime.datetime
    quantity: float

    exchange_name:str



"""
Tools for visualizing dependencies between Terms.
"""

from contextlib import contextmanager
import errno
from functools import partial
from io import BytesIO
from subprocess import Popen, PIPE

from networkx import topological_sort

from ziplime.pipeline.data import BoundColumn
from ziplime.pipeline import Filter, Factor, Classifier, Term
from ziplime.pipeline.term import AssetExists


class NoIPython(Exception):
    pass


def delimit(delimiters, content):
    """
    Surround `content` with the first and last characters of `delimiters`.

    >>> delimit('[]', "foo")  # doctest: +SKIP
    '[foo]'
    >>> delimit('""', "foo")  # doctest: +SKIP
    '"foo"'
    """
    if len(delimiters) != 2:
        raise ValueError("`delimiters` must be of length 2. Got %r" % delimiters)
    return "".join([delimiters[0], content, delimiters[1]])


quote = partial(delimit, '""')
bracket = partial(delimit, "[]")


def begin_graph(f, name, **attrs):
    writeln(f, "strict digraph %s {" % name)
    writeln(f, "graph {}".format(format_attrs(attrs)))


def begin_cluster(f, name, **attrs):
    attrs.setdefault("label", quote(name))
    writeln(f, "subgraph cluster_%s {" % name)
    writeln(f, "graph {}".format(format_attrs(attrs)))


def end_graph(f):
    writeln(f, "}")


@contextmanager
def graph(f, name, **attrs):
    begin_graph(f, name, **attrs)
    yield
    end_graph(f)


@contextmanager
def cluster(f, name, **attrs):
    begin_cluster(f, name, **attrs)
    yield
    end_graph(f)


def roots(g):
    "Get nodes from graph G with indegree 0"
    return set(n for n, d in g.in_degree().items() if d == 0)


def filter_nodes(include_asset_exists, nodes):
    if include_asset_exists:
        return nodes
    return filter(lambda n: n is not AssetExists(), nodes)


def _render(g, out, format_, include_asset_exists=False):
    """
    Draw `g` as a graph to `out`, in format `format`.

    Parameters
    ----------
    g : ziplime.pipeline.graph.TermGraph
        Graph to render.
    out : file-like object
    format_ : str {'png', 'svg'}
        Output format.
    include_asset_exists : bool
        Whether to filter out `AssetExists()` nodes.
    """
    graph_attrs = {"rankdir": "TB", "splines": "ortho"}
    cluster_attrs = {"style": "filled", "color": "lightgoldenrod1"}

    in_nodes = g.loadable_terms
    out_nodes = list(g.outputs.values())

    f = BytesIO()
    with graph(f, "G", **graph_attrs):

        # Write outputs cluster.
        with cluster(f, "Output", labelloc="b", **cluster_attrs):
            for term in filter_nodes(include_asset_exists, out_nodes):
                add_term_node(f, term)

        # Write inputs cluster.
        with cluster(f, "Input", **cluster_attrs):
            for term in filter_nodes(include_asset_exists, in_nodes):
                add_term_node(f, term)

        # Write intermediate results.
        for term in filter_nodes(include_asset_exists, topological_sort(g.graph)):
            if term in in_nodes or term in out_nodes:
                continue
            add_term_node(f, term)

        # Write edges
        for source, dest in g.graph.edges():
            if source is AssetExists() and not include_asset_exists:
                continue
            add_edge(f, id(source), id(dest))

    cmd = ["dot", "-T", format_]
    try:
        proc = Popen(cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE)
    except OSError as exc:
        if exc.errno == errno.ENOENT:
            raise RuntimeError(
                "Couldn't find `dot` graph layout program. "
                "Make sure Graphviz is installed and `dot` is on your path."
            ) from exc
        else:
            raise

    f.seek(0)
    proc_stdout, proc_stderr = proc.communicate(f.read())
    if proc_stderr:
        raise RuntimeError(
            "Error(s) while rendering graph: %s" % proc_stderr.decode("utf-8")
        )

    out.write(proc_stdout)


def display_graph(g, format="svg", include_asset_exists=False):
    """
    Display a TermGraph interactively from within IPython.
    """
    try:
        import IPython.display as display
    except ImportError as exc:
        raise NoIPython("IPython is not installed.  Can't display graph.") from exc

    if format == "svg":
        display_cls = display.SVG
    elif format in ("jpeg", "png"):
        display_cls = partial(display.Image, format=format, embed=True)

    out = BytesIO()
    _render(g, out, format, include_asset_exists=include_asset_exists)
    return display_cls(data=out.getvalue())


def writeln(f, s):
    f.write((s + "\n").encode("utf-8"))


def fmt(obj):
    if isinstance(obj, Term):
        r = obj.graph_repr()
    else:
        r = obj
    return '"%s"' % r


def add_term_node(f, term):
    declare_node(f, id(term), attrs_for_node(term))


def declare_node(f, name, attributes):
    writeln(f, "{0} {1};".format(name, format_attrs(attributes)))


def add_edge(f, source, dest):
    writeln(f, "{0} -> {1};".format(source, dest))


def attrs_for_node(term, **overrides):
    attrs = {
        "shape": "box",
        "colorscheme": "pastel19",
        "style": "filled",
        "label": fmt(term),
    }
    if isinstance(term, BoundColumn):
        attrs["fillcolor"] = "1"
    if isinstance(term, Factor):
        attrs["fillcolor"] = "2"
    elif isinstance(term, Filter):
        attrs["fillcolor"] = "3"
    elif isinstance(term, Classifier):
        attrs["fillcolor"] = "4"

    attrs.update(**overrides or {})
    return attrs


def format_attrs(attrs):
    """
    Format key, value pairs from attrs into graphviz attrs format

    Examples
    --------
    >>> format_attrs({'key1': 'value1', 'key2': 'value2'})  # doctest: +SKIP
    '[key1=value1, key2=value2]'
    """
    if not attrs:
        return ""
    entries = ["=".join((key, value)) for key, value in attrs.items()]
    return "[" + ", ".join(entries) + "]"



import datetime
import importlib.util
import sys
import traceback
import uuid
from collections import namedtuple, OrderedDict
from contextlib import AsyncExitStack
from copy import copy
import warnings
from typing import Callable
import pandas as pd
import structlog

import ziplime
from itertools import chain, repeat

from exchange_calendars import ExchangeCalendar

from ziplime.assets.domain.asset_type import AssetType
from ziplime.assets.entities.asset import Asset
from ziplime.assets.services.asset_service import AssetService
from ziplime.constants.logging_event import LoggingEvent, LoggingEvent
from ziplime.core.algorithm_file import AlgorithmFile
from ziplime.data.services.data_source import DataSource
from ziplime.domain.bar_data import BarData
from ziplime.finance.blotter.blotter import Blotter
from ziplime.finance.controls.long_only import LongOnly
from ziplime.finance.controls.max_order_count import MaxOrderCount
from ziplime.finance.controls.max_order_size import MaxOrderSize
from ziplime.finance.controls.max_position_size import MaxPositionSize
from ziplime.finance.controls.min_leverage import MinLeverage
from ziplime.finance.controls.restricted_list_order import RestrictedListOrder
from ziplime.finance.domain.ledger import Ledger
from ziplime.finance.domain.order import Order
from ziplime.finance.domain.order_status import OrderStatus
from ziplime.finance.metrics import MaxLeverage
from ziplime.gens.domain.trading_clock import TradingClock
from ziplime.exchanges.exchange import Exchange
from ziplime.trading.base_trading_algorithm import BaseTradingAlgorithm
from ziplime.trading.entities.orders.limit_order_request import LimitOrderRequest
from ziplime.trading.entities.orders.market_order_request import MarketOrderRequest
from ziplime.trading.entities.orders.order_request import OrderRequest
from ziplime.trading.entities.trading_pair import TradingPair
from ziplime.trading.enums.order_side import OrderSide
from ziplime.trading.enums.order_type import OrderType
from ziplime.trading.enums.simulation_event import SimulationEvent
from ziplime.trading.trading_signal_executor import TradingSignalExecutor
from ziplime.utils.calendar_utils import get_calendar

from ziplime.protocol import handle_non_market_minutes
from ziplime.errors import (
    AttachPipelineAfterInitialize,
    CannotOrderDelistedAsset,
    DuplicatePipelineName,
    IncompatibleCommissionModel,
    IncompatibleSlippageModel,
    NoSuchPipeline,
    OrderDuringInitialize,
    OrderInBeforeTradingStart,
    PipelineOutputDuringInitialize,
    RegisterAccountControlPostInit,
    RegisterTradingControlPostInit,
    ScheduleFunctionInvalidCalendar,
    SetCancelPolicyPostInit,
    SetCommissionPostInit,
    SetSlippagePostInit,
    UnsupportedCancelPolicy,
    UnsupportedDatetimeFormat,
    ZeroCapitalError, SymbolNotFound, BarSimulationError,
)

from ziplime.finance.execution import ExecutionStyle
from ziplime.finance.asset_restrictions import Restrictions
from ziplime.finance.cancel_policy import CancelPolicy
from ziplime.finance.asset_restrictions import (
    NoRestrictions,
)
from ziplime.assets.entities.futures_contract import FuturesContract
from ziplime.assets.entities.equity import Equity
from ziplime.finance.domain.simulation_paremeters import SimulationParameters
from ziplime.finance.metrics_tracker import MetricsTracker
from ziplime.pipeline import Pipeline
import ziplime.pipeline.domain as domain
from ziplime.pipeline.engine import (
    ExplodingPipelineEngine,
    SimplePipelineEngine,
)
from ziplime.utils.api_support import (
    api_method,
    require_initialized,
    require_not_initialized,
    ZiplineAPI,
    disallowed_in_before_trading_start,
)
from ziplime.utils.compat import ExitStack
from ziplime.utils.date_utils import make_utc_aware
from ziplime.utils.cache import ExpiringCache

from ziplime.utils.events import (
    EventManager,
    make_eventrule,
    date_rules,
    time_rules,
    calendars,
    AfterOpen,
    BeforeClose, EventRule,
)
from ziplime.utils.math_utils import (
    tolerant_equals,
    round_if_near_integer,
)
from ziplime.sources.benchmark_source import BenchmarkSource

# For creating and storing pipeline instances
AttachedPipeline = namedtuple("AttachedPipeline", "pipe chunks eager")


class NoBenchmark(ValueError):
    def __init__(self):
        super(NoBenchmark, self).__init__(
            "Must specify either benchmark_sid or benchmark_returns.",
        )


class TradingAlgorithm(BaseTradingAlgorithm):
    """A class that represents a trading strategy and parameters to execute
    the strategy.

    Parameters
    ----------
    *args, **kwargs
        Forwarded to ``initialize`` unless listed below.
    initialize : callable[context -> None], optional
        Function that is called at the start of the simulation to
        setup the initial context.
    handle_data : callable[(context, data) -> None], optional
        Function called on every bar. This is where most logic should be
        implemented.
    before_trading_start : callable[(context, data) -> None], optional
        Function that is called before any bars have been processed each
        day.
    analyze : callable[(context, DataFrame) -> None], optional
        Function that is called at the end of the backtest. This is passed
        the context and the performance results for the backtest.
    script : str, optional
        Algoscript that contains the definitions for the four algorithm
        lifecycle functions and any supporting code.
    namespace : dict, optional
        The namespace to execute the algoscript in. By default this is an
        empty namespace that will include only python built ins.
    algo_filename : str, optional
        The filename for the algoscript. This will be used in exception
        tracebacks. default: '<string>'.
    data_frequency : {'daily', 'minute'}, optional
        The duration of the bars.
    equities_metadata : dict or DataFrame or file-like object, optional
        If dict is provided, it must have the following structure:
        * keys are the identifiers
        * values are dicts containing the metadata, with the metadata
          field name as the key
        If pandas.DataFrame is provided, it must have the
        following structure:
        * column names must be the metadata fields
        * index must be the different asset identifiers
        * array contents should be the metadata value
        If an object with a ``read`` method is provided, ``read`` must
        return rows containing at least one of 'sid' or 'symbol' along
        with the other metadata fields.
    futures_metadata : dict or DataFrame or file-like object, optional
        The same layout as ``equities_metadata`` except that it is used
        for futures information.
    identifiers : list, optional
        Any asset identifiers that are not provided in the
        equities_metadata, but will be traded by this TradingAlgorithm.
    get_pipeline_loader : callable[BoundColumn -> PipelineLoader], optional
        The function that maps pipeline columns to their loaders.
    create_event_context : callable[BarData -> context manager], optional
        A function used to create a context mananger that wraps the
        execution of all events that are scheduled for a bar.
        This function will be passed the data for the bar and should
        return the actual context manager that will be entered.
    history_container_class : type, optional
        The type of history container to use. default: HistoryContainer
    adjustment_reader : AdjustmentReader
        The interface to the adjustments.
    """

    def __init__(
            self,
            algorithm: AlgorithmFile,
            asset_service: AssetService,
            exchanges: dict[str, Exchange],
            # Algorithm API
            metrics_set,
            blotter: Blotter,
            benchmark_source: BenchmarkSource | None,
            clock: TradingClock,
            custom_data_sources: list[DataSource],
            capital_changes=None,
            get_pipeline_loader=None,
            create_event_context=None,
            stop_on_error: bool = False
    ):
        self.algorithm = algorithm
        self.exchanges = exchanges
        self.stop_on_error = stop_on_error
        self.default_exchange = list(self.exchanges.values())[0]
        self.trading_signal_executor = TradingSignalExecutor()
        # List of trading controls to be used to validate orders.
        self.trading_controls = []

        # List of account controls to be checked on each bar.
        self.account_controls = []

        self._recorded_vars = {}
        self.namespace = {}

        # XXX: This is kind of a mess.
        # We support passing a data_portal in `run`, but we need an asset
        # finder earlier than that to look up assets for things like
        # set_benchmark.
        # self.data_portal = data_portal
        # self.market_data_provider = market_data_provider
        self.benchmark_source = benchmark_source

        # XXX: This is also a mess. We should remove all of this and only allow
        #      one way to pass a calendar.
        #
        # self.sim_params = sim_params
        self.asset_service = asset_service
        self.clock = clock

        self.metrics_tracker = None
        self._last_sync_time = pd.NaT
        self._metrics_set = metrics_set

        # Initialize Pipeline API data.
        self.init_engine(get_pipeline_loader)
        self._pipelines = {}

        # Create an already-expired cache so that we compute the first time
        # data is requested.
        self._pipeline_cache = ExpiringCache()

        self.blotter = blotter
        self.new_orders = OrderedDict()
        # The symbol lookup date specifies the date to use when resolving
        # symbols to sids, and can be set using set_symbol_lookup_date()
        self._symbol_lookup_date = None

        self._initialize = None
        self._before_trading_start = None
        self._analyze = None

        self._in_before_trading_start = False

        self.event_manager = EventManager(create_event_context)

        self._handle_data = None
        self._ledger = Ledger(trading_sessions=clock.sessions, exchanges=exchanges,
                              data_frequency=clock.emission_rate)

        self._initialize = algorithm.initialize
        self._handle_data = algorithm.handle_data
        self._before_trading_start = algorithm.before_trading_start
        # Optional analyze function, gets called after run
        self._analyze = algorithm.analyze

        self.event_manager.add_event(
            ziplime.utils.events.Event(
                ziplime.utils.events.Always(),
                # We pass handle_data.__func__ to get the unbound method.
                # We will explicitly pass the algorithm to bind it again.
                self.handle_data.__func__,
            ),
            prepend=True,
        )
        data_sources = {}

        # TODO: what if we add funds?
        for exchange in exchanges.values():
            data_sources[exchange.name] = exchange
            if exchange.get_start_cash_balance() <= 0:
                raise ZeroCapitalError()

        # Prepare the algo for initialization
        self.initialized = False

        # A dictionary of capital changes, keyed by timestamp, indicating the
        # target/delta of the capital changes, along with values
        self.capital_changes = capital_changes or {}

        # A dictionary of the actual capital change deltas, keyed by timestamp
        self.capital_change_deltas = {}

        self.restrictions = NoRestrictions()
        for ds in custom_data_sources:
            data_sources[ds.name] = ds
        self.current_data = BarData(
            simulation_dt_func=self.get_datetime,
            trading_calendar=self.clock.trading_calendar,
            restrictions=self.restrictions,
            data_sources=data_sources
        )

        # We don't have a datetime for the current snapshot until we
        # receive a message.
        self.simulation_dt = None

        self.clock = clock

        self._logger = structlog.get_logger(__name__)

    def init_engine(self, get_loader):
        """Construct and store a PipelineEngine from loader.

        If get_loader is None, constructs an ExplodingPipelineEngine
        """
        if get_loader is not None:
            self.engine = SimplePipelineEngine(
                get_loader,
                self.asset_service,
                self.default_pipeline_domain(self.clock.trading_calendar),
            )
        else:
            self.engine = ExplodingPipelineEngine()

    async def initialize(self, *args, **kwargs):
        """Call self._initialize with `self` made available to Zipline API
        functions.
        """
        with ZiplineAPI(self):
            await self._initialize(self, *args, **kwargs)

    def before_trading_start(self, data):
        self.compute_eager_pipelines()

        if self._before_trading_start is None:
            return

        self._in_before_trading_start = True

        with handle_non_market_minutes(
                data
        ) if self.clock.emission_rate == datetime.timedelta(minutes=1) else ExitStack():
            self._before_trading_start(self, data)

        self._in_before_trading_start = False

    async def handle_data(self, data):
        if self._handle_data:
            await self._handle_data(self, data)

    def analyze(self, perf):
        if self._analyze is None:
            return

        with ZiplineAPI(self):
            self._analyze(self, perf)

    def compute_eager_pipelines(self):
        """Compute any pipelines attached with eager=True."""
        for name, pipe in self._pipelines.items():
            if pipe.eager:
                self.pipeline_output(name)

    async def get_generator(self):
        """Override this method to add new logic to the construction
        of the generator. Overrides can use the _create_generator
        method to get a standard construction generator.
        """
        self.metrics_tracker = MetricsTracker(
            asset_service=self.asset_service,
            exchanges=self.exchanges,
            trading_calendar=self.clock.trading_calendar,
            sessions=self.clock.sessions,
            first_session=self.clock.start_session,
            last_session=self.clock.end_session,
            emission_rate=self.clock.emission_rate,
            ledger=self._ledger,
            metrics=self._metrics_set,
            benchmark_source=self.benchmark_source
        )

        # Set the dt initially to the period start by forcing it to change.
        # self.on_dt_changed(dt=self.clock.start_session)
        # self.datetime =self.clock.start_session
        self.simulation_dt = self.clock.start_session
        if not self.initialized:
            await self.initialize()
            self.initialized = True
        self.metrics_tracker.handle_start_of_simulation()
        return self.transform()

    async def run(self):
        """Run the algorithm."""
        self._logger.info("Running algorithm")
        # HACK: I don't think we really want to support passing a data portal
        # this late in the long term, but this is needed for now for backwards
        # compat downstream.

        # Create ziplime and loop through simulated_trading.
        # Each iteration returns a perf dictionary
        try:
            perfs = []
            errors = []
            async for perf, errors in await self.get_generator():
                perfs.append(perf)

            # convert perf dict to pandas dataframe
            daily_stats = self._create_daily_stats(perfs)

            self.analyze(daily_stats)
        finally:
            self.data_portal = None
            self.metrics_tracker = None

        return daily_stats, errors

    def _create_daily_stats(self, perfs):
        # create daily and cumulative stats dataframe
        daily_perfs = []
        # TODO: the loop here could overwrite expected properties
        # of daily_perf. Could potentially raise or log a
        # warning.
        for perf in perfs:
            if "daily_perf" in perf:
                perf["daily_perf"].update(perf["daily_perf"].pop("recorded_vars"))
                perf["daily_perf"].update(perf["cumulative_risk_metrics"])
                daily_perfs.append(perf["daily_perf"])
            else:
                self.risk_report = perf

        daily_dts = pd.DatetimeIndex([p["period_close"] for p in daily_perfs])
        daily_dts = make_utc_aware(daily_dts)
        daily_stats = pd.DataFrame(daily_perfs, index=daily_dts)
        return daily_stats

    def calculate_capital_changes(
            self, dt: datetime.datetime, emission_rate: datetime.timedelta, is_interday: bool,
            portfolio_value_adjustment: float = 0.00
    ):
        """If there is a capital change for a given dt, this means the the change
        occurs before `handle_data` on the given dt. In the case of the
        change being a target value, the change will be computed on the
        portfolio value according to prices at the given dt

        `portfolio_value_adjustment`, if specified, will be removed from the
        portfolio_value of the cumulative performance when calculating deltas
        from target capital changes.
        """

        # CHECK is try/catch faster than search?

        try:
            capital_change = self.capital_changes[dt]
        except KeyError:
            return

        self._sync_last_sale_prices()
        if capital_change["type"] == "target":
            target = capital_change["value"]
            capital_change_amount = target - (
                    self.portfolio.portfolio_value - portfolio_value_adjustment
            )

            self._logger.info(
                "Processing capital change to target %s at %s. Capital "
                "change delta is %s" % (target, dt, capital_change_amount)
            )
        elif capital_change["type"] == "delta":
            target = None
            capital_change_amount = capital_change["value"]
            self._logger.info(
                "Processing capital change of delta %s at %s"
                % (capital_change_amount, dt)
            )
        else:
            self._logger.error(
                "Capital change %s does not indicate a valid type "
                "('target' or 'delta')" % capital_change
            )
            return

        self.capital_change_deltas.update({dt: capital_change_amount})
        self._ledger.capital_change(change_amount=capital_change_amount)

        yield {
            "capital_change": {
                "date": dt,
                "type": "cash",
                "target": target,
                "delta": capital_change_amount,
            }
        }

    def add_event(self, rule: EventRule, callback: Callable):
        """Adds an event to the algorithm's EventManager.

        Parameters
        ----------
        rule : EventRule
            The rule for when the callback should be triggered.
        callback : callable[(context, data) -> None]
            The function to execute when the rule is triggered.
        """
        self.event_manager.add_event(
            event=ziplime.utils.events.Event(rule=rule, callback=callback),
        )

    @api_method
    def schedule_function(
            self,
            func: Callable,
            date_rule: EventRule = None,
            time_rule: EventRule = None,
            half_days: bool = True,
            calendar: ExchangeCalendar | None = None,
    ):
        """Schedule a function to be called repeatedly in the future.

        Parameters
        ----------
        func : callable
            The function to execute when the rule is triggered. ``func`` should
            have the same signature as ``handle_data``.
        date_rule : ziplime.utils.events.EventRule, optional
            Rule for the dates on which to execute ``func``. If not
            passed, the function will run every trading day.
        time_rule : ziplime.utils.events.EventRule, optional
            Rule for the time at which to execute ``func``. If not passed, the
            function will execute at the end of the first market minute of the
            day.
        half_days : bool, optional
            Should this rule fire on half days? Default is True.
        calendar : Sentinel, optional
            Calendar used to compute rules that depend on the trading calendar.

        See Also
        --------
        :class:`ziplime.api.date_rules`
        :class:`ziplime.api.time_rules`
        """

        # When the user calls schedule_function(func, <time_rule>), assume that
        # the user meant to specify a time rule but no date rule, instead of
        # a date rule and no time rule as the signature suggests
        if isinstance(date_rule, (AfterOpen, BeforeClose)) and not time_rule:
            warnings.warn(
                "Got a time rule for the second positional argument "
                "date_rule. You should use keyword argument "
                "time_rule= when calling schedule_function without "
                "specifying a date_rule",
                stacklevel=3,
            )

        date_rule = date_rule or date_rules.every_day()
        time_rule = (
            (time_rule or time_rules.every_minute())
            if self.clock.emission_rate == datetime.timedelta(minutes=1)
            else
            # If we are in daily mode the time_rule is ignored.
            time_rules.every_minute()
        )

        # Check the type of the algorithm's schedule before pulling calendar
        # Note that the ExchangeTradingSchedule is currently the only
        # TradingSchedule class, so this is unlikely to be hit
        if calendar is None:
            cal = self.clock.trading_calendar
        elif calendar is calendars.US_EQUITIES:
            cal = get_calendar("XNYS")
        elif calendar is calendars.US_FUTURES:
            cal = get_calendar("us_futures")
        else:
            raise ScheduleFunctionInvalidCalendar(
                given_calendar=calendar,
                allowed_calendars="[calendars.US_EQUITIES, calendars.US_FUTURES]",
            )

        self.add_event(
            rule=make_eventrule(date_rule=date_rule, time_rule=time_rule, cal=cal, half_days=half_days),
            callback=func,
        )

    @api_method
    def record(self, *args, **kwargs):
        """Track and record values each day.

        Parameters
        ----------
        **kwargs
            The names and values to record.

        Notes
        -----
        These values will appear in the performance packets and the performance
        dataframe passed to ``analyze`` and returned from
        :func:`~ziplime.run_algorithm`.
        """
        # Make 2 objects both referencing the same iterator
        args = [iter(args)] * 2

        # Zip generates list entries by calling `next` on each iterator it
        # receives.  In this case the two iterators are the same object, so the
        # call to next on args[0] will also advance args[1], resulting in zip
        # returning (a,b) (c,d) (e,f) rather than (a,a) (b,b) (c,c) etc.
        positionals = zip(*args)
        for name, value in chain(positionals, kwargs.items()):
            self._recorded_vars[name] = value

    @api_method
    def continuous_future(
            self, root_symbol_str: str, offset: int = 0, roll: str = "volume", adjustment: str = "mul"
    ):
        """Create a specifier for a continuous contract.

        Parameters
        ----------
        root_symbol_str : str
            The root symbol for the future chain.

        offset : int, optional
            The distance from the primary contract. Default is 0.

        roll_style : str, optional
            How rolls are determined. Default is 'volume'.

        adjustment : str, optional
            Method for adjusting lookback prices between rolls. Options are
            'mul', 'add', and None. Default is 'mul'.

        Returns
        -------
        continuous_future : ziplime.assets.ContinuousFuture
            The continuous future specifier.
        """
        return self.data_portal._data_bundle.asset_repository.create_continuous_future(
            root_symbol_str,
            offset,
            roll,
            adjustment,
        )

    @api_method
    async def symbol(
            self, symbol_str: str, asset_type: AssetType = AssetType.EQUITY,
            exchange_name: str = None,
            country_code: str | None = None
    ) -> Asset | None:
        """Lookup an Equity by its ticker symbol.

        Parameters
        ----------
        symbol_str : str
            The ticker symbol for the equity to lookup.
        country_code : str or None, optional
            A country to limit symbol searches to.

        Returns
        -------
        equity : ziplime.assets.Equity
            The equity that held the ticker symbol on the current
            symbol lookup date.

        Raises
        ------
        SymbolNotFound
            Raised when the symbols was not held on the current lookup date.

        See Also
        --------
        :func:`ziplime.api.set_symbol_lookup_date`
        """
        # If the user has not set the symbol lookup date,
        # use the end_session as the date for symbol->sid resolution.
        # _lookup_date = (
        #     self._symbol_lookup_date
        #     if self._symbol_lookup_date is not None
        #     else pd.Timestamp(self.sim_params.end_session).to_pydatetime().date()
        # )
        if exchange_name is None:
            exchange_name = self.default_exchange.name

        asset = await self.asset_service.get_asset_by_symbol(symbol=symbol_str,
                                                             asset_type=asset_type,
                                                             exchange_name=exchange_name)
        if asset is None:
            raise SymbolNotFound(symbol=symbol_str)
        return asset

    @api_method
    def symbols(self, *args, **kwargs):
        """Lookup multuple Equities as a list.

        Parameters
        ----------
        *args : iterable[str]
            The ticker symbols to lookup.
        country_code : str or None, optional
            A country to limit symbol searches to.

        Returns
        -------
        equities : list[ziplime.assets.Equity]
            The equities that held the given ticker symbols on the current
            symbol lookup date.

        Raises
        ------
        SymbolNotFound
            Raised when one of the symbols was not held on the current
            lookup date.

        See Also
        --------
        :func:`ziplime.api.set_symbol_lookup_date`
        """
        return [self.symbol(identifier, **kwargs) for identifier in args]

    @api_method
    async def sid(self, sid: int) -> Asset | None:
        """Lookup an Asset by its unique asset identifier.

        Parameters
        ----------
        sid : int
            The unique integer that identifies an asset.

        Returns
        -------
        asset : ziplime.assets.Asset
            The asset with the given ``sid``.

        Raises
        ------
        SidsNotFound
            When a requested ``sid`` does not map to any asset.
        """
        return await self.asset_service.get_asset_by_sid(sid=sid)

    @api_method
    async def future_symbol(self, symbol: str, exchange_name: str = None) -> FuturesContract | None:
        """Lookup a futures contract with a given symbol.

        Parameters
        ----------
        symbol : str
            The symbol of the desired contract.

        Returns
        -------
        future : ziplime.assets.Future
            The future that trades with the name ``symbol``.

        Raises
        ------
        SymbolNotFound
            Raised when no contract named 'symbol' is found.
        """
        return await self.asset_service.get_futures_contract_by_symbol(
            symbol=symbol,
            exchange_name=exchange_name or self.default_exchange.name
        )

    def _calculate_order_value_amount(self, asset: Asset, value: float, exchange_name: str):
        """Calculates how many shares/contracts to order based on the type of
        asset being ordered.
        """
        # Make sure the asset exists, and that there is a last price for it.
        # FIXME: we should use BarData's can_trade logic here, but I haven't
        # yet found a good way to do that.
        normalized_date = self.clock.trading_calendar.minute_to_session(self.simulation_dt).date()

        if normalized_date < asset.start_date:
            raise CannotOrderDelistedAsset(
                msg=f"Cannot order sid={asset.sid}, as it started trading on {asset.start_date}"
            )
        elif normalized_date > asset.end_date:
            raise CannotOrderDelistedAsset(
                msg=f"Cannot order sid={asset.sid}, as it stopped trading on {asset.end_date}."
            )
        else:
            # last_price = self.current_data.current([asset], fields={"price"})["price"][0]
            last_price_data = self.exchanges[exchange_name].current(frozenset({asset}),
                                                               dt=self.simulation_dt,
                                                               fields=frozenset({"price"}))["price"]
            if len(last_price_data) == 0:
            # if last_price is None:
                raise CannotOrderDelistedAsset(
                    msg=f"Cannot order sid={asset.sid} on {self.simulation_dt} as there is no last price for the security."
                )
            last_price = last_price_data[0]
            if last_price is None:
                raise CannotOrderDelistedAsset(
                    msg=f"Cannot order sid={asset.sid} on {self.simulation_dt} as there is no last price for the security."
                )
        if tolerant_equals(last_price, 0):
            self._logger.debug(f"Price of 0 for {asset}; can't infer value")
            # Don't place any order
            return 0
        if type(asset) is FuturesContract:
            return value / (last_price * asset.multiplier)
        else:
            return value / last_price

    def _can_order_asset(self, asset: Asset):
        if asset.auto_close_date:
            day = self.clock.trading_calendar.minute_to_session(self.simulation_dt).date()

            if day > min(asset.end_date, asset.auto_close_date):
                # If we are after the asset's end date or auto close date, warn
                # the user that they can't place an order for this asset, and
                # return None.
                self._logger.warning(
                    f"Cannot place order for sid={asset.sid}, as it has de-listed. "
                    f"Any existing positions for this asset will be "
                    f"liquidated on "
                    f"{asset.auto_close_date}."
                )

                return False

        return True

    def reject_order(self, order_id: str, reason: str = ""):
        """
        Mark the given order as 'rejected', which is functionally similar to
        cancelled. The distinction is that rejections are involuntary (and
        usually include a message from a exchange indicating why the order was
        rejected) while cancels are typically user-driven.
        """
        order = self.blotter.get_order_by_id(order_id)
        if order is None:
            return
        order.reject(reason=reason)
        order.dt = self.simulation_dt

        self.blotter.order_rejected(order=order)
        # we want this order's new status to be relayed out
        # along with newly placed orders.
        self.new_orders.move_to_end(order_id)

    def hold_order(self, order_id: str, reason: str = ""):
        """
        Mark the order with order_id as 'held'. Held is functionally similar
        to 'open'. When a fill (full or partial) arrives, the status
        will automatically change back to open/filled as necessary.
        """
        order = self.blotter.get_order_by_id(order_id)
        if order is None or not order.open:
            return
        order.hold(reason=reason)
        order.dt = self.simulation_dt
        # we want this order's new status to be relayed out
        # along with newly placed orders.
        self.new_orders.move_to_end(order.id)

    @api_method
    @disallowed_in_before_trading_start(OrderInBeforeTradingStart())
    async def order(self, asset: Asset, amount: int, style: ExecutionStyle,
                    exchange_name: str | None = None) -> Order | None:
        """Place an order for a fixed number of shares.

        Parameters
        ----------
        asset : Asset
            The asset to be ordered.
        amount : int
            The amount of shares to order. If ``amount`` is positive, this is
            the number of shares to buy or cover. If ``amount`` is negative,
            this is the number of shares to sell or short.
        style : ExecutionStyle, optional
            The execution style for the order.

        Returns
        -------
        order_id : str or None
            The unique identifier for this order, or None if no order was
            placed.

        Notes
        -----
        The ``limit_price`` and ``stop_price`` arguments provide shorthands for
        passing common execution styles. Passing ``limit_price=N`` is
        equivalent to ``style=LimitOrder(N)``. Similarly, passing
        ``stop_price=M`` is equivalent to ``style=StopOrder(M)``, and passing
        ``limit_price=N`` and ``stop_price=M`` is equivalent to
        ``style=StopLimitOrder(N, M)``. It is an error to pass both a ``style``
        and ``limit_price`` or ``stop_price``.

        See Also
        --------
        :class:`ziplime.finance.execution.ExecutionStyle`
        :func:`ziplime.api.order_value`
        :func:`ziplime.api.order_percent`
        """
        if not self._can_order_asset(asset=asset):
            return None
        # TODO: implement dynamic risk control

        self.validate_order_params(asset=asset, amount=amount)
        if exchange_name is None:
            exchange = self.default_exchange
        else:
            exchange = self.exchanges[exchange_name]
        order_id = None
        order_qty_rounded = int(round_if_near_integer(amount))

        order = Order(
            dt=self.simulation_dt,
            asset=asset,
            amount=order_qty_rounded,
            id=order_id,
            commission=0.00,
            filled=0,
            execution_style=style,
            status=OrderStatus.OPEN,
            exchange_name=exchange.name
        )
        if amount == 0:
            self._logger.warning("Not executing order for zero shares.")
            return None

        quote_asset = await self.asset_service.get_currency_by_symbol(symbol="USD",
                                                                      exchange_name=exchange_name)

        self._logger.info(LoggingEvent.ORDER_SUBMIT, style=str(style), quantity=order_qty_rounded,
                          symbol=asset.get_symbol_by_exchange(exchange_name),
                          simulation_dt=self.simulation_dt)

        submitted_order = await self.default_exchange.submit_order(order=order)
        # quote_asset = await self.asset_service.get_currency_by_symbol(symbol="USD",
        #                                                               exchange_name=exchange_name)

        # match style.to_order_type():
        #     case OrderType.LIMIT:
        #         order_req = LimitOrderRequest(
        #             order_id=uuid.uuid4().hex,
        #             trading_pair=TradingPair(base_asset=asset,
        #                                      quote_asset=quote_asset,
        #                                      exchange_name=exchange.name),
        #             order_side=OrderSide.BUY if amount > 0 else OrderSide.SELL,
        #             quantity=float(amount),
        #             limit_price=style.get_limit_price(is_buy=amount > 0)
        #         )
        #     case OrderType.MARKET:
        #         order_req = MarketOrderRequest(
        #             order_id=uuid.uuid4().hex,
        #             trading_pair=TradingPair(base_asset=asset,
        #                                      quote_asset=quote_asset,
        #                                      exchange_name=exchange.name),
        #             order_side=OrderSide.BUY if amount > 0 else OrderSide.SELL,
        #             quantity=float(amount),
        #             exchange_name=exchange.name,
        #             creation_date=self.simulation_dt
        #         )
        #     case _:
        #         raise ValueError(f"Unsupported order type: {style.to_order_type()}")

        # trading_signal = await self.trading_signal_executor.create_order_execute_trading_signal(
        #     algorithm=self,
        #     order=order_req,
        #     exchange=exchange
        # )

        persisted_order = self.blotter.save_order(order=submitted_order)
        self.new_orders[submitted_order.id] = submitted_order

        return submitted_order

    def new_order_submitted(self, order: Order):
        self.blotter.save_order(order=order)
        self.new_orders[order.id] = order
        return order

    def validate_order_params(self, asset: Asset, amount: int):
        """
        Helper method for validating parameters to the order API function.

        Raises an UnsupportedOrderParameters if invalid arguments are found.
        """

        if not self.initialized:
            raise OrderDuringInitialize(
                msg="order() can only be called from within handle_data()"
            )

        for control in self.trading_controls:
            control.validate(
                asset=asset,
                amount=amount,
                portfolio=self.portfolio,
                algo_datetime=self.simulation_dt,
                algo_current_data=self.current_data,
            )

    @api_method
    @disallowed_in_before_trading_start(OrderInBeforeTradingStart())
    async def order_value(self, asset: Asset, value: float, limit_price: float | None = None,
                          stop_price: float | None = None,
                          style: ExecutionStyle | None = None,
                          exchange_name: str | None = None
                          ):
        """Place an order for a fixed amount of money.

        Equivalent to ``order(asset, value / data.current(asset, 'price'))``.

        Parameters
        ----------
        asset : Asset
            The asset to be ordered.
        value : float
            Amount of value of ``asset`` to be transacted. The number of shares
            bought or sold will be equal to ``value / current_price``.
        limit_price : float, optional
            Limit price for the order.
        stop_price : float, optional
            Stop price for the order.
        style : ExecutionStyle
            The execution style for the order.

        Returns
        -------
        order_id : str
            The unique identifier for this order.

        Notes
        -----
        See :func:`ziplime.api.order` for more information about
        ``limit_price``, ``stop_price``, and ``style``

        See Also
        --------
        :class:`ziplime.finance.execution.ExecutionStyle`
        :func:`ziplime.api.order`
        :func:`ziplime.api.order_percent`
        """
        if not self._can_order_asset(asset):
            return None

        amount = self._calculate_order_value_amount(asset=asset, value=value,
                                                    exchange_name=exchange_name or self.default_exchange.name)
        return await self.order(
            asset,
            amount,
            limit_price=limit_price,
            stop_price=stop_price,
            style=style,
            exchange_name=exchange_name
        )

    @property
    def recorded_vars(self):
        return copy(self._recorded_vars)

    def _sync_last_sale_prices(self, dt: datetime.datetime = None):
        """Sync the last sale prices on the metrics tracker to a given
        datetime.

        Parameters
        ----------
        dt : datetime
            The time to sync the prices to.

        Notes
        -----
        This call is cached by the datetime. Repeated calls in the same bar
        are cheap.
        """
        if dt is None:
            dt = self.simulation_dt

        if dt != self._last_sync_time:
            self._ledger.sync_last_sale_prices(dt=dt, handle_non_market_minutes=False)
            self._last_sync_time = dt

    @property
    def portfolio(self):
        self._sync_last_sale_prices()
        return self._ledger.portfolio

    @property
    def account(self):
        self._sync_last_sale_prices()
        return self._ledger.account

    # def on_dt_changed(self, dt):
    #     """Callback triggered by the simulation loop whenever the current dt
    #     changes.
    #
    #     Any logic that should happen exactly once at the start of each datetime
    #     group should happen here.
    #     """
    #     self.datetime = dt

    @api_method
    def get_datetime(self):
        """Returns the current simulation datetime.

        Parameters
        ----------
        tz : tzinfo or str, optional
            The timezone to return the datetime in. This defaults to utc.

        Returns
        -------
        dt : datetime
            The current simulation datetime converted to ``tz``.
        """
        return self.simulation_dt
        # dt = self.datetime
        # return dt

    @api_method
    def set_slippage(self, us_equities=None, us_futures=None):
        """Set the slippage models for the simulation.

        Parameters
        ----------
        us_equities : EquitySlippageModel
            The slippage model to use for trading US equities.
        us_futures : FutureSlippageModel
            The slippage model to use for trading US futures.

        Notes
        -----
        This function can only be called during
        :func:`~ziplime.api.initialize`.

        See Also
        --------
        :class:`ziplime.finance.slippage.SlippageModel`
        """
        if self.initialized:
            raise SetSlippagePostInit()

        if us_equities is not None:
            if Equity not in us_equities.allowed_asset_types:
                raise IncompatibleSlippageModel(
                    asset_type="equities",
                    given_model=us_equities,
                    supported_asset_types=us_equities.allowed_asset_types,
                )
            self.blotter.slippage_models[Equity] = us_equities

        if us_futures is not None:
            if FuturesContract not in us_futures.allowed_asset_types:
                raise IncompatibleSlippageModel(
                    asset_type="futures",
                    given_model=us_futures,
                    supported_asset_types=us_futures.allowed_asset_types,
                )
            self.blotter.slippage_models[FuturesContract] = us_futures

    @api_method
    def set_commission(self, us_equities=None, us_futures=None):
        """Sets the commission models for the simulation.

        Parameters
        ----------
        us_equities : EquityCommissionModel
            The commission model to use for trading US equities.
        us_futures : FutureCommissionModel
            The commission model to use for trading US futures.

        Notes
        -----
        This function can only be called during
        :func:`~ziplime.api.initialize`.

        See Also
        --------
        :class:`ziplime.finance.commission.PerShare`
        :class:`ziplime.finance.commission.PerTrade`
        :class:`ziplime.finance.commission.PerDollar`
        """
        if self.initialized:
            raise SetCommissionPostInit()

        if us_equities is not None:
            if Equity not in us_equities.allowed_asset_types:
                raise IncompatibleCommissionModel(
                    asset_type="equities",
                    given_model=us_equities,
                    supported_asset_types=us_equities.allowed_asset_types,
                )
            self.blotter.commission_models[Equity] = us_equities

        if us_futures is not None:
            if FuturesContract not in us_futures.allowed_asset_types:
                raise IncompatibleCommissionModel(
                    asset_type="futures",
                    given_model=us_futures,
                    supported_asset_types=us_futures.allowed_asset_types,
                )
            self.blotter.commission_models[FuturesContract] = us_futures

    @api_method
    def set_cancel_policy(self, cancel_policy):
        """Sets the order cancellation policy for the simulation.

        Parameters
        ----------
        cancel_policy : CancelPolicy
            The cancellation policy to use.

        See Also
        --------
        :class:`ziplime.api.EODCancel`
        :class:`ziplime.api.NeverCancel`
        """
        if not isinstance(cancel_policy, CancelPolicy):
            raise UnsupportedCancelPolicy()

        if self.initialized:
            raise SetCancelPolicyPostInit()

        self.blotter.cancel_policy = cancel_policy

    @api_method
    def set_symbol_lookup_date(self, dt):
        """Set the date for which symbols will be resolved to their assets
        (symbols may map to different firms or underlying assets at
        different times)

        Parameters
        ----------
        dt : datetime
            The new symbol lookup date.
        """
        try:
            self._symbol_lookup_date = pd.Timestamp(dt).tz_localize("UTC")
        except TypeError:
            self._symbol_lookup_date = pd.Timestamp(dt).tz_convert("UTC")
        except ValueError as exc:
            raise UnsupportedDatetimeFormat(
                input=dt, method="set_symbol_lookup_date"
            ) from exc

    # @property
    # def data_frequency(self):
    #     return self.sim_params.data_frequency
    #
    # @data_frequency.setter
    # def data_frequency(self, value):
    #     assert value in ("daily", "minute")
    #     self.sim_params.data_frequency = value

    @api_method
    @disallowed_in_before_trading_start(OrderInBeforeTradingStart())
    async def order_percent(
            self, asset: Asset, percent: float, style: ExecutionStyle,
            exchange_name: str | None = None
    ):
        """Place an order in the specified asset corresponding to the given
        percent of the current portfolio value.

        Parameters
        ----------
        asset : Asset
            The asset that this order is for.
        percent : float
            The percentage of the portfolio value to allocate to ``asset``.
            This is specified as a decimal, for example: 0.50 means 50%.
        style : ExecutionStyle
            The execution style for the order.

        Returns
        -------
        order_id : str
            The unique identifier for this order.

        Notes
        -----
        See :func:`ziplime.api.order` for more information about
        ``limit_price``, ``stop_price``, and ``style``

        See Also
        --------
        :class:`ziplime.finance.execution.ExecutionStyle`
        :func:`ziplime.api.order`
        :func:`ziplime.api.order_value`
        """
        if not self._can_order_asset(asset=asset):
            return None

        amount = self._calculate_order_percent_amount(asset=asset, percent=percent,
                                                      exchange_name=exchange_name or self.default_exchange.name)
        return await self.order(
            asset=asset,
            amount=amount,
            style=style,
            exchange_name=exchange_name
        )

    def _calculate_order_percent_amount(self, asset: Asset, percent: float, exchange_name: str):
        value = self.portfolio.portfolio_value * percent
        return self._calculate_order_value_amount(asset=asset, value=value, exchange_name=exchange_name)

    @api_method
    @disallowed_in_before_trading_start(OrderInBeforeTradingStart())
    async def order_target(
            self, asset: Asset, target: int, style: ExecutionStyle,
            exchange_name: str | None = None
    ):
        """Place an order to adjust a position to a target number of shares. If
        the position doesn't already exist, this is equivalent to placing a new
        order. If the position does exist, this is equivalent to placing an
        order for the difference between the target number of shares and the
        current number of shares.

        Parameters
        ----------
        asset : Asset
            The asset that this order is for.
        target : int
            The desired number of shares of ``asset``.
        limit_price : float, optional
            The limit price for the order.
        stop_price : float, optional
            The stop price for the order.
        style : ExecutionStyle
            The execution style for the order.

        Returns
        -------
        order_id : str
            The unique identifier for this order.


        Notes
        -----
        ``order_target`` does not take into account any open orders. For
        example:

        .. code-block:: python

           order_target(sid(0), 10)
           order_target(sid(0), 10)

        This code will result in 20 shares of ``sid(0)`` because the first
        call to ``order_target`` will not have been filled when the second
        ``order_target`` call is made.

        See :func:`ziplime.api.order` for more information about
        ``limit_price``, ``stop_price``, and ``style``

        See Also
        --------
        :class:`ziplime.finance.execution.ExecutionStyle`
        :func:`ziplime.api.order`
        :func:`ziplime.api.order_target_percent`
        :func:`ziplime.api.order_target_value`
        """
        if not self._can_order_asset(asset=asset):
            return None

        amount = self._calculate_order_target_amount(asset=asset, target=target)
        return await self.order(
            asset=asset,
            amount=amount,
            style=style,
            exchange_name=exchange_name
        )

    def _calculate_order_target_amount(self, asset: Asset, target: int):
        if asset in self.portfolio.positions:
            current_position = self.portfolio.positions[asset].amount
            target -= current_position

        return target

    @api_method
    @disallowed_in_before_trading_start(OrderInBeforeTradingStart())
    async def order_target_value(
            self, asset: Asset, target: float, style: ExecutionStyle,
            exchange_name: str | None = None
    ):
        """Place an order to adjust a position to a target value. If
        the position doesn't already exist, this is equivalent to placing a new
        order. If the position does exist, this is equivalent to placing an
        order for the difference between the target value and the
        current value.
        If the Asset being ordered is a Future, the 'target value' calculated
        is actually the target exposure, as Futures have no 'value'.

        Parameters
        ----------
        asset : Asset
            The asset that this order is for.
        target : float
            The desired total value of ``asset``.
        style : ExecutionStyle
            The execution style for the order.

        Returns
        -------
        order_id : str
            The unique identifier for this order.

        Notes
        -----
        ``order_target_value`` does not take into account any open orders. For
        example:

        .. code-block:: python

           order_target_value(sid(0), 10)
           order_target_value(sid(0), 10)

        This code will result in 20 dollars of ``sid(0)`` because the first
        call to ``order_target_value`` will not have been filled when the
        second ``order_target_value`` call is made.

        See Also
        --------
        :class:`ziplime.finance.execution.ExecutionStyle`
        :func:`ziplime.api.order`
        :func:`ziplime.api.order_target`
        :func:`ziplime.api.order_target_percent`
        """
        if not self._can_order_asset(asset):
            return None

        target_amount = self._calculate_order_value_amount(asset=asset, value=target,
                                                           exchange_name=exchange_name or self.default_exchange.name)
        amount = self._calculate_order_target_amount(asset, target_amount)
        return await self.order(
            asset=asset,
            amount=amount,
            style=style,
            exchange_name=exchange_name
        )

    @api_method
    @disallowed_in_before_trading_start(OrderInBeforeTradingStart())
    async def order_target_percent(
            self, asset: Asset, target: float,
            style: ExecutionStyle, exchange_name: str | None = None
    ):
        """Place an order to adjust a position to a target percent of the
        current portfolio value. If the position doesn't already exist, this is
        equivalent to placing a new order. If the position does exist, this is
        equivalent to placing an order for the difference between the target
        percent and the current percent.

        Parameters
        ----------
        asset : Asset
            The asset that this order is for.
        target : float
            The desired percentage of the portfolio value to allocate to
            ``asset``. This is specified as a decimal, for example:
            0.50 means 50%.
        style : ExecutionStyle
            The execution style for the order.

        Returns
        -------
        order_id : str
            The unique identifier for this order.

        Notes
        -----
        ``order_target_value`` does not take into account any open orders. For
        example:

        .. code-block:: python

           order_target_percent(sid(0), 10)
           order_target_percent(sid(0), 10)

        This code will result in 20% of the portfolio being allocated to sid(0)
        because the first call to ``order_target_percent`` will not have been
        filled when the second ``order_target_percent`` call is made.

        See :func:`ziplime.api.order` for more information about
        ``limit_price``, ``stop_price``, and ``style``

        See Also
        --------
        :class:`ziplime.finance.execution.ExecutionStyle`
        :func:`ziplime.api.order`
        :func:`ziplime.api.order_target`
        :func:`ziplime.api.order_target_value`
        """
        if not (-1 <= target <= 1):
            raise ValueError("target must be between -1 and 1")
        if not self._can_order_asset(asset):
            return None

        target_amount = self._calculate_order_percent_amount(asset=asset, percent=target,
                                                             exchange_name=exchange_name or self.default_exchange.name)
        amount = self._calculate_order_target_amount(asset=asset, target=target_amount)

        return await self.order(
            asset=asset,
            amount=amount,
            style=style,
            exchange_name=exchange_name
        )

    # def _calculate_order_target_percent_amount(self, asset, target):
    #     target_amount = self._calculate_order_percent_amount(asset, target)
    #     return self._calculate_order_target_amount(asset, target_amount)

    # @api_method
    # def batch_market_order(self, share_counts):
    #     """Place a batch market order for multiple assets.
    #
    #     Parameters
    #     ----------
    #     share_counts : pd.Series[Asset -> int]
    #         Map from asset to number of shares to order for that asset.
    #
    #     Returns
    #     -------
    #     order_ids : pd.Index[str]
    #         Index of ids for newly-created orders.
    #     """
    #     style = MarketOrder()
    #     order_args = [
    #         (asset, amount, style) for (asset, amount) in share_counts.items() if amount
    #     ]
    #     return self.blotter.batch_order(order_args)

    @api_method
    def get_open_orders(self, asset=None):
        """Retrieve all of the current open orders.

        Parameters
        ----------
        asset : Asset
            If passed and not None, return only the open orders for the given
            asset instead of all open orders.

        Returns
        -------
        open_orders : dict[list[Order]] or list[Order]
            If no asset is passed this will return a dict mapping Assets
            to a list containing all the open orders for the asset.
            If an asset is passed then this will return a list of the open
            orders for this asset.
        """
        if asset is None:
            return {
                key: list(orders.values())  # [order.to_api_obj() for order in orders]
                for key, orders in self.blotter.open_orders.items()
                if orders
            }
        if asset in self.blotter.open_orders:
            orders = self.blotter.open_orders[asset]
            return list(orders.values())
            # return [order.to_api_obj() for order in orders]
        return []

    @api_method
    def get_order(self, order_id) -> Order | None:
        """Lookup an order based on the order id returned from one of the
        order functions.

        Parameters
        ----------
        order_id : str
            The unique identifier for the order.

        Returns
        -------
        order : Order
            The order object.
        """
        self.exchange.get_orders_by_ids([order_id])
        return self.blotter.get_order_by_id(order_id=order_id)

    @api_method
    def cancel_order(self, order_id: str, relay_status: bool = True) -> None:
        """Cancel an open order.

        Parameters
        ----------
        order_param : str or Order
            The order_id or order object to cancel.
        """
        order = self.blotter.get_order_by_id(order_id=order_id)
        if order is None or not order.open:
            return
        order.cancel()
        order.dt = self.simulation_dt
        # we want this order's new status to be relayed out
        # along with newly placed orders.

        self.blotter.order_cancelled(order=order)
        self.exchange.cancel_order(order_id=order_id)
        if relay_status:
            self.new_orders[order.id] = order
        else:
            self.new_orders.pop(order.id, None)

    def cancel_all_orders_for_asset(self, asset: Asset, warn: bool = False, relay_status: bool = True):
        """
        Cancel all open orders for a given asset.
        """
        # (sadly) open_orders is a defaultdict, so this will always succeed.
        orders = self.blotter.get_open_orders_by_asset(asset=asset)
        if not orders:
            return
        # We're making a copy here because `cancel` mutates the list of open
        # orders in place.  The right thing to do here would be to make
        # self.open_orders no longer a defaultdict.  If we do that, then we
        # should just remove the orders once here and be done with the matter.
        for order_id, order in orders.items():
            self.cancel_order(order_id=order.id, relay_status=relay_status)
            if warn:
                # Message appropriately depending on whether there's
                # been a partial fill or not.
                if order.filled > 0:
                    self._logger.warning(
                        f"Your order for {order.amount} shares of "
                        f"{order.asset.sid} has been partially filled. "
                        f"{order.filled} shares were successfully "
                        f"purchased. {order.amount - order.filled} shares were not "
                        f"filled by the end of day and "
                        f"were canceled."
                    )
                elif order.filled < 0:
                    self._logger.warning(
                        f"Your order for {order.amount} shares of "
                        f"{asset.sid} has been partially filled. "
                        f"{-1 * order.filled} shares were successfully "
                        f"sold. {-1 * (order.amount - order.filled)} shares were not "
                        f"filled by the end of day and "
                        f"were canceled."
                    )
                else:
                    self._logger.warning(
                        f"Your order for {order.amount} shares of "
                        f"{order.asset.sid} failed to fill by the end of day "
                        f"and was canceled."
                    )
        self.blotter.cancel_all_orders_for_asset(asset=asset, relay_status=relay_status)

    ####################
    # Account Controls #
    ####################

    def register_account_control(self, control):
        """
        Register a new AccountControl to be checked on each bar.
        """
        if self.initialized:
            raise RegisterAccountControlPostInit()
        self.account_controls.append(control)

    def validate_account_controls(self):
        for control in self.account_controls:
            control.validate(
                self.portfolio,
                self.account,
                self.simulation_dt,
                self.current_data,
            )

    @api_method
    def set_max_leverage(self, max_leverage):
        """Set a limit on the maximum leverage of the algorithm.

        Parameters
        ----------
        max_leverage : float
            The maximum leverage for the algorithm. If not provided there will
            be no maximum.
        """
        control = MaxLeverage(max_leverage)
        self.register_account_control(control)

    @api_method
    def set_min_leverage(self, min_leverage, grace_period):
        """Set a limit on the minimum leverage of the algorithm.

        Parameters
        ----------
        min_leverage : float
            The minimum leverage for the algorithm.
        grace_period : pd.Timedelta
            The offset from the start date used to enforce a minimum leverage.
        """
        deadline = self.sim_params.start_session + grace_period
        control = MinLeverage(min_leverage, deadline)
        self.register_account_control(control)

    ####################
    # Trading Controls #
    ####################

    def register_trading_control(self, control):
        """
        Register a new TradingControl to be checked prior to order calls.
        """
        if self.initialized:
            raise RegisterTradingControlPostInit()
        self.trading_controls.append(control)

    @api_method
    def set_max_position_size(
            self, asset=None, max_shares=None, max_notional=None, on_error="fail"
    ):
        """Set a limit on the number of shares and/or dollar value held for the
        given sid. Limits are treated as absolute values and are enforced at
        the time that the algo attempts to place an order for sid. This means
        that it's possible to end up with more than the max number of shares
        due to splits/dividends, and more than the max notional due to price
        improvement.

        If an algorithm attempts to place an order that would result in
        increasing the absolute value of shares/dollar value exceeding one of
        these limits, raise a TradingControlException.

        Parameters
        ----------
        asset : Asset, optional
            If provided, this sets the guard only on positions in the given
            asset.
        max_shares : int, optional
            The maximum number of shares to hold for an asset.
        max_notional : float, optional
            The maximum value to hold for an asset.
        """
        control = MaxPositionSize(
            asset=asset,
            max_shares=max_shares,
            max_notional=max_notional,
            on_error=on_error,
        )
        self.register_trading_control(control)

    @api_method
    def set_max_order_size(
            self, asset=None, max_shares=None, max_notional=None, on_error="fail"
    ):
        """Set a limit on the number of shares and/or dollar value of any single
        order placed for sid.  Limits are treated as absolute values and are
        enforced at the time that the algo attempts to place an order for sid.

        If an algorithm attempts to place an order that would result in
        exceeding one of these limits, raise a TradingControlException.

        Parameters
        ----------
        asset : Asset, optional
            If provided, this sets the guard only on positions in the given
            asset.
        max_shares : int, optional
            The maximum number of shares that can be ordered at one time.
        max_notional : float, optional
            The maximum value that can be ordered at one time.
        """
        control = MaxOrderSize(
            asset=asset,
            max_shares=max_shares,
            max_notional=max_notional,
            on_error=on_error,
        )
        self.register_trading_control(control)

    @api_method
    def set_max_order_count(self, max_count, on_error="fail"):
        """Set a limit on the number of orders that can be placed in a single
        day.

        Parameters
        ----------
        max_count : int
            The maximum number of orders that can be placed on any single day.
        """
        control = MaxOrderCount(on_error, max_count)
        self.register_trading_control(control)

    @api_method
    def set_asset_restrictions(self, restrictions: Restrictions, on_error: str = "fail"):
        """Set a restriction on which assets can be ordered.

        Parameters
        ----------
        restricted_list : Restrictions
            An object providing information about restricted assets.

        See Also
        --------
        ziplime.finance.asset_restrictions.Restrictions
        """
        control = RestrictedListOrder(on_error, restrictions)
        self.register_trading_control(control)
        self.restrictions |= restrictions

    @api_method
    def set_long_only(self, on_error="fail"):
        """Set a rule specifying that this algorithm cannot take short
        positions.
        """
        self.register_trading_control(LongOnly(on_error))

    ##############
    # Pipeline API
    ##############
    @api_method
    @require_not_initialized(AttachPipelineAfterInitialize())
    def attach_pipeline(self, pipeline, name, chunks=None, eager=True):
        """Register a pipeline to be computed at the start of each day.

        Parameters
        ----------
        pipeline : Pipeline
            The pipeline to have computed.
        name : str
            The name of the pipeline.
        chunks : int or iterator, optional
            The number of days to compute pipeline results for. Increasing
            this number will make it longer to get the first results but
            may improve the total runtime of the simulation. If an iterator
            is passed, we will run in chunks based on values of the iterator.
            Default is True.
        eager : bool, optional
            Whether or not to compute this pipeline prior to
            before_trading_start.

        Returns
        -------
        pipeline : Pipeline
            Returns the pipeline that was attached unchanged.

        See Also
        --------
        :func:`ziplime.api.pipeline_output`
        """
        if chunks is None:
            # Make the first chunk smaller to get more immediate results:
            # (one week, then every half year)
            chunks = chain([5], repeat(126))
        elif isinstance(chunks, int):
            chunks = repeat(chunks)

        if name in self._pipelines:
            raise DuplicatePipelineName(name=name)

        self._pipelines[name] = AttachedPipeline(pipeline, iter(chunks), eager)

        # Return the pipeline to allow expressions like
        # p = attach_pipeline(Pipeline(), 'name')
        return pipeline

    @api_method
    @require_initialized(PipelineOutputDuringInitialize())
    def pipeline_output(self, name):
        """Get results of the pipeline attached by with name ``name``.

        Parameters
        ----------
        name : str
            Name of the pipeline from which to fetch results.

        Returns
        -------
        results : pd.DataFrame
            DataFrame containing the results of the requested pipeline for
            the current simulation date.

        Raises
        ------
        NoSuchPipeline
            Raised when no pipeline with the name `name` has been registered.

        See Also
        --------
        :func:`ziplime.api.attach_pipeline`
        :meth:`ziplime.pipeline.engine.PipelineEngine.run_pipeline`
        """
        try:
            pipe, chunks, _ = self._pipelines[name]
        except KeyError as exc:
            raise NoSuchPipeline(
                name=name,
                valid=list(self._pipelines.keys()),
            ) from exc
        return self._pipeline_output(pipe, chunks, name)

    def _pipeline_output(self, pipeline, chunks, name):
        """Internal implementation of `pipeline_output`."""
        # TODO FIXME TZ MESS
        today = self.simulation_dt
        try:
            data = self._pipeline_cache.get(key=name, dt=today)
        except KeyError:
            # Calculate the next block.
            data, valid_until = self.run_pipeline(
                pipeline=pipeline,
                start_session=today,
                chunksize=next(chunks),
            )
            self._pipeline_cache.set(key=name, value=data, expiration_dt=valid_until)

        # Now that we have a cached result, try to return the data for today.
        try:
            return data.loc[today]
        except KeyError:
            # This happens if no assets passed the pipeline screen on a given
            # day.
            return pd.DataFrame(index=[], columns=data.columns)

    def run_pipeline(self, pipeline, start_session, chunksize):
        """Compute `pipeline`, providing values for at least `start_date`.

        Produces a DataFrame containing data for days between `start_date` and
        `end_date`, where `end_date` is defined by:

            `end_date = min(start_date + chunksize trading days,
                            simulation_end)`

        Returns
        -------
        (data, valid_until) : tuple (pd.DataFrame, datetime.datetime)

        See Also
        --------
        PipelineEngine.run_pipeline
        """
        sessions = self.clock.trading_calendar.sessions

        # Load data starting from the previous trading day...
        start_date_loc = sessions.get_loc(start_session)

        # ...continuing until either the day before the simulation end, or
        # until chunksize days of data have been loaded.
        sim_end_session = self.sim_params.end_session

        end_loc = min(start_date_loc + chunksize, sessions.get_loc(sim_end_session))

        end_session = sessions[end_loc]

        return (
            self.engine.run_pipeline(pipeline, start_session, end_session),
            end_session,
        )

    @staticmethod
    def default_pipeline_domain(calendar):
        """Get a default pipeline domain for algorithms running on ``calendar``.

        This will be used to infer a domain for pipelines that only use generic
        datasets when running in the context of a TradingAlgorithm.
        """
        return _DEFAULT_DOMAINS.get(calendar.name, domain.GENERIC)

    ##################
    # End Pipeline API
    ##################

    # def get_simulation_dt(self) -> datetime.datetime:
    #     return self.simulation_dt

    def execute_order_cancellation_policy(self):
        self.blotter.execute_cancel_policy(SimulationEvent.SESSION_END)

    def calculate_minute_capital_changes(self, dt: datetime.datetime):
        # process any capital changes that came between the last
        # and current minutes
        return self.calculate_capital_changes(dt, emission_rate=self.metrics_tracker.emission_rate,
                                              is_interday=False)

    # TODO: simplify
    # flake8: noqa: C901
    async def every_bar(
            self,
            dt_to_use: datetime.datetime,
            current_data: BarData,
            handle_data,
    ):
        # print(f"dt_to_use: in every_bar: {dt_to_use}")
        for capital_change in self.calculate_minute_capital_changes(dt_to_use):
            yield capital_change

        self.simulation_dt = dt_to_use
        # self.datetime = dt_to_use
        # called every tick (minute or day).
        # self.on_dt_changed(dt=dt_to_use)

        # handle any transactions and commissions coming out new orders
        # placed in the last bar
        new_transactions = []
        new_commissions = []
        closed_orders = []
        for exchange in self.exchanges.values():
            (
                new_trans,
                new_comm,
                closed,
            ) = await exchange.get_transactions(
                orders=self.blotter.get_open_orders(exchange_name=exchange.name),
                current_dt=self.simulation_dt
            )
            new_transactions.extend(new_trans)
            new_commissions.extend(new_comm)
            closed_orders.extend(closed)
        # print(f"getting transactions for {current_data.current_dt}, new transactions: {len(new_transactions)}, new commissions: {len(new_commissions)}, closed orders: {len(closed_orders)}" )
        self.blotter.prune_orders(closed_orders=closed_orders)

        for transaction in new_transactions:
            self._ledger.process_transaction(transaction=transaction)

            if transaction.order_id is None:
                # TODO: fix this when we get back order id in transaction
                continue

            # since this order was modified, record it
            order = self.blotter.get_order_by_id(transaction.order_id, exchange_name=transaction.exchange_name)
            self._ledger.process_order(order=order)

        for commission in new_commissions:
            self._ledger.process_commission(commission=commission)

        await handle_data(context=self, data=current_data, dt=dt_to_use)

        # grab any new orders from the blotter, then clear the list.
        # this includes cancelled orders.
        new_orders = self.new_orders
        self.new_orders = dict()

        # if we have any new orders, record them so that we know
        # in what perf period they were placed.
        for new_order in new_orders.values():
            self._ledger.process_order(order=new_order)

    def once_a_day(
            self,
            midnight_dt,
            current_data,
            asset_service,
    ):
        # process any capital changes that came overnight
        for capital_change in self.calculate_capital_changes(
                midnight_dt, emission_rate=self.metrics_tracker.emission_rate,
                is_interday=True
        ):
            yield capital_change

        # set all the timestamps
        self.simulation_dt = midnight_dt
        # self.datetime = midnight_dt
        # self.on_dt_changed(midnight_dt)

        self.metrics_tracker.handle_market_open(session_label=midnight_dt)

        # handle any splits that impact any positions or any open orders.
        assets_we_care_about = (
                self._ledger.position_tracker.positions.keys() | self.blotter.get_all_assets_in_open_orders()
        )

        if assets_we_care_about:
            splits = asset_service.get_splits(assets_we_care_about, midnight_dt)
            if splits:
                self.blotter.process_splits(splits)
                self._ledger.process_splits(splits)

    def on_exit(self):
        # Remove references to algo, data portal, et al to break cycles
        # and ensure deterministic cleanup of these objects when the
        # simulation finishes.
        self.benchmark_source = self.current_data = None

    async def transform(self):
        """
        Main generator work loop.
        """

        async with (AsyncExitStack() as stack):
            stack.callback(self.on_exit)
            stack.enter_context(ZiplineAPI(algo_instance=self))

            # if self.data_frequency < datetime.timedelta(days=1):
            #
            #     def execute_order_cancellation_policy():
            #         self.blotter.execute_cancel_policy(SimulationEvent.SESSION_END)
            #
            #     def calculate_minute_capital_changes(dt: datetime.datetime):
            #         # process any capital changes that came between the last
            #         # and current minutes
            #         return self.calculate_capital_changes(dt, emission_rate=emission_rate, is_interday=False)
            #
            # elif self.data_frequency == datetime.timedelta(days=1):
            #
            #     def execute_order_cancellation_policy():
            #         self.blotter.execute_daily_cancel_policy(SimulationEvent.SESSION_END)
            #
            #     def calculate_minute_capital_changes(dt: datetime.datetime):
            #         return []
            #
            # else:
            #
            #     def execute_order_cancellation_policy():
            #         pass
            #
            #     def calculate_minute_capital_changes(dt: datetime.datetime):
            #         return []
            errors = []
            for dt, action in self.clock:
                try:
                    if action == SimulationEvent.BAR:
                        async for capital_change_packet in self.every_bar(dt_to_use=dt, current_data=self.current_data,
                                                                          handle_data=self.event_manager.handle_data):
                            yield capital_change_packet, []
                    elif action == SimulationEvent.SESSION_START:
                        for capital_change_packet in self.once_a_day(midnight_dt=dt,
                                                                     current_data=self.current_data,
                                                                     asset_service=self.asset_service):
                            yield capital_change_packet, []
                    elif action == SimulationEvent.SESSION_END:
                        # End of the session.
                        positions = self._ledger.position_tracker.positions
                        position_assets = list(positions.keys())

                        # await self.asset_service.retrieve_all(
                        #     sids=[a.sid for a in positions]
                        # )

                        self._cleanup_expired_assets(dt=dt, position_assets=position_assets)

                        self.execute_order_cancellation_policy()
                        self.validate_account_controls()

                        yield self._get_daily_message(dt=dt), []
                    elif action == SimulationEvent.BEFORE_TRADING_START_BAR:
                        self.simulation_dt = dt
                        # self.datetime = dt
                        # self.on_dt_changed(dt=dt)
                        self.before_trading_start(data=self.current_data)
                    elif action == SimulationEvent.EMISSION_RATE_END and self.clock.emission_rate == datetime.timedelta(
                            minutes=1):
                        minute_msg = self._get_minute_message(
                            dt=dt,
                        )

                        yield minute_msg, []
                except Exception as e:
                    errors.append(
                        BarSimulationError(
                            trace=traceback.format_exc(),
                            message=f"Exception raised during simulation dt={dt}: {e}",
                            simulation_dt=dt
                        )
                    )
                    self._logger.error(f"Simulation error on dt={dt}")
                    if self.stop_on_error:
                        raise
            risk_message = self.metrics_tracker.handle_simulation_end()
            yield risk_message, errors

    def _cleanup_expired_assets(self, dt: datetime.datetime, position_assets):
        """
        Clear out any assets that have expired before starting a new sim day.

        Performs two functions:

        1. Finds all assets for which we have open orders and clears any
           orders whose assets are on or after their auto_close_date.

        2. Finds all assets for which we have positions and generates
           close_position events for any assets that have reached their
           auto_close_date.
        """

        def past_auto_close_date(asset: Asset):
            acd = asset.auto_close_date
            if acd is not None:
                acd = acd
            return acd is not None and acd <= dt.date()

        # Remove positions in any sids that have reached their auto_close date.
        assets_to_clear = [
            asset
            for asset in position_assets
            if past_auto_close_date(asset)
        ]
        # data_portal = self.data_portal
        for asset in assets_to_clear:
            self._ledger.close_position(asset=asset, dt=dt)

        # Remove open orders for any sids that have reached their auto close
        # date. These orders get processed immediately because otherwise they
        # would not be processed until the first bar of the next day.
        assets_to_cancel = [
            asset
            for asset in self.blotter.get_all_assets_in_open_orders()
            if past_auto_close_date(asset=asset)
        ]

        for asset in assets_to_cancel:
            self.cancel_all_orders_for_asset(asset=asset)

        # Make a copy here so that we are not modifying the list that is being
        # iterated over.
        new_order_values = list(self.new_orders.values())
        # print(self.new_orders)
        for order in new_order_values:
            if order.status == OrderStatus.CANCELLED:
                self._ledger.process_order(order=order)
                self.new_orders.pop(order.id)

    def _get_daily_message(self, dt: datetime.datetime):
        """
        Get a perf message for the given datetime.
        """
        perf_message = self.metrics_tracker.handle_market_close(
            dt=dt,
        )
        perf_message["daily_perf"]["recorded_vars"] = self.recorded_vars
        return perf_message

    def _get_minute_message(self, dt: datetime.datetime):
        """
        Get a perf message for the given datetime.
        """
        rvars = self.recorded_vars

        minute_message = self.metrics_tracker.handle_minute_close(
            dt=dt,
        )

        minute_message["minute_perf"]["recorded_vars"] = rvars
        return minute_message


# Map from calendar name to default domain for that calendar.
_DEFAULT_DOMAINS = {d.calendar_name: d for d in domain.BUILT_IN_DOMAINS}



import datetime
import os
from pathlib import Path

import asyncio
from exchange_calendars import ExchangeCalendar
from ziplime.utils.calendar_utils import get_calendar

from ziplime.assets.domain.ordered_contracts import CHAIN_PREDICATES
from ziplime.assets.entities.currency import Currency
from ziplime.assets.entities.currency_symbol_mapping import CurrencySymbolMapping
from ziplime.assets.entities.equity import Equity
from ziplime.assets.entities.equity_symbol_mapping import EquitySymbolMapping
from ziplime.assets.entities.symbol_universe import SymbolsUniverse
from ziplime.assets.models.exchange_info import ExchangeInfo
from ziplime.assets.repositories.sqlalchemy_adjustments_repository import SqlAlchemyAdjustmentRepository
from ziplime.assets.repositories.sqlalchemy_asset_repository import SqlAlchemyAssetRepository
from ziplime.assets.services.asset_service import AssetService
from ziplime.constants.period import Period
from ziplime.constants.stock_symbols import ALL_US_STOCK_SYMBOLS
from ziplime.data.data_sources.asset_data_source import AssetDataSource
from ziplime.data.services.bundle_service import BundleService
from ziplime.data.services.data_bundle_source import DataBundleSource
from ziplime.data.services.file_system_bundle_registry import FileSystemBundleRegistry
from ziplime.data.services.file_system_parquet_bundle_storage import FileSystemParquetBundleStorage


def get_asset_service(db_path: str = str(Path(Path.home(), ".ziplime", "assets.sqlite").absolute()),
                      clear_asset_db: bool = False) -> AssetService:
    if clear_asset_db and os.path.exists(db_path):
        os.remove(db_path)
    db_url = f"sqlite+aiosqlite:///{db_path}"
    assets_repository = SqlAlchemyAssetRepository(db_url=db_url, future_chain_predicates=CHAIN_PREDICATES)
    adjustments_repository = SqlAlchemyAdjustmentRepository(db_url=db_url)
    asset_service = AssetService(asset_repository=assets_repository, adjustments_repository=adjustments_repository)
    return asset_service

async def ingest_assets(asset_service: AssetService, asset_data_source: AssetDataSource):
    asset_start_date = datetime.datetime(year=1900, month=1, day=1, tzinfo=datetime.timezone.utc)
    asset_end_date = datetime.datetime(year=2099, month=1, day=1, tzinfo=datetime.timezone.utc)

    usd_currency = Currency(
        asset_name="USD",
        symbol_mapping={
            "LIME": CurrencySymbolMapping(
                symbol="USD",
                exchange_name="LIME",
                start_date=asset_start_date,
                end_date=asset_end_date
            )
        },
        sid=None,
        start_date=asset_start_date,
        end_date=asset_end_date,
        auto_close_date=asset_end_date,
        first_traded=asset_start_date,
        mic=None,
    )
    assets = await asset_data_source.get_assets()
    equities = [
        Equity(
            asset_name=asset["symbol"],
            symbol_mapping={
                "LIME": EquitySymbolMapping(
                    symbol=asset["symbol"],
                    exchange_name="LIME",
                    start_date=asset_start_date,
                    end_date=asset_end_date,
                    company_symbol="",
                    share_class_symbol=""
                )
            },
            sid=None,
            start_date=asset_start_date,
            end_date=asset_end_date,
            auto_close_date=asset_end_date,
            first_traded=asset_start_date,
            mic=asset["mic"]
        ) for asset in assets.iter_rows(named=True)

    ]

    await asset_service.save_exchanges(
        exchanges=[ExchangeInfo(exchange="LIME", canonical_name="LIME", country_code="US")])
    await asset_service.save_currencies([usd_currency])
    await asset_service.save_equities(equities)


async def ingest_default_assets(asset_service: AssetService, asset_data_source: AssetDataSource):
    asset_start_date = datetime.datetime(year=1900, month=1, day=1, tzinfo=datetime.timezone.utc)
    asset_end_date = datetime.datetime(year=2099, month=1, day=1, tzinfo=datetime.timezone.utc)

    usd_currency = Currency(
        asset_name="USD",
        symbol_mapping={
            "LIME": CurrencySymbolMapping(
                symbol="USD",
                exchange_name="LIME",
                start_date=asset_start_date,
                end_date=asset_end_date
            )
        },
        sid=None,
        start_date=asset_start_date,
        end_date=asset_end_date,
        auto_close_date=asset_end_date,
        first_traded=asset_start_date,
        mic=None
    )

    equities = [
        Equity(
            asset_name=symbol,
            symbol_mapping={
                "LIME": EquitySymbolMapping(
                    symbol=symbol,
                    exchange_name="LIME",
                    start_date=asset_start_date,
                    end_date=asset_end_date,
                    company_symbol="",
                    share_class_symbol=""
                )
            },
            sid=None,
            start_date=asset_start_date,
            end_date=asset_end_date,
            auto_close_date=asset_end_date,
            first_traded=asset_start_date,
            mic=None
        ) for symbol in ALL_US_STOCK_SYMBOLS

    ]

    await asset_service.save_exchanges(
        exchanges=[ExchangeInfo(exchange="LIME", canonical_name="LIME", country_code="US")])
    await asset_service.save_currencies([usd_currency])
    await asset_service.save_equities(equities)


async def ingest_symbol_universes(asset_service: AssetService, asset_data_source: AssetDataSource):
    sp500 = await asset_data_source.get_constituents(index='SP500')
    assets_symbols = list(sp500[0])

    equities = await asset_service.get_equities_by_symbols(symbols=assets_symbols)
    universe = SymbolsUniverse(
        name="SP500",
        universe_type="index",
        symbol="SP500",
        assets=equities
    )
    await asset_service.save_symbol_universe(symbol_universe=universe)


async def _ingest_market_data(
        start_date: datetime.datetime,
        end_date: datetime.datetime,
        trading_calendar: ExchangeCalendar,
        bundle_name: str,
        symbols: list[str],
        frequency: datetime.timedelta,
        data_bundle_source: DataBundleSource,
        asset_service: AssetService,
        forward_fill_missing_ohlcv_data: bool = True,
        bundle_storage_path: str = str(Path(Path.home(), ".ziplime", "data")),
):
    bundle_registry = FileSystemBundleRegistry(base_data_path=bundle_storage_path)
    bundle_service = BundleService(bundle_registry=bundle_registry)
    bundle_storage = FileSystemParquetBundleStorage(base_data_path=bundle_storage_path, compression_level=5)

    bundle_version = str(int(datetime.datetime.now(tz=trading_calendar.tz).timestamp()))

    await bundle_service.ingest_market_data_bundle(
        date_start=start_date.replace(tzinfo=trading_calendar.tz),
        date_end=end_date.replace(tzinfo=trading_calendar.tz),
        bundle_storage=bundle_storage,
        data_bundle_source=data_bundle_source,
        frequency=frequency,
        symbols=symbols,
        name=bundle_name,
        bundle_version=bundle_version,
        trading_calendar=trading_calendar,
        asset_service=asset_service,
        forward_fill_missing_ohlcv_data=forward_fill_missing_ohlcv_data,
    )


async def _ingest_custom_data(
        start_date: datetime.datetime,
        end_date: datetime.datetime,
        trading_calendar: ExchangeCalendar,
        bundle_name: str,
        symbols: list[str],
        frequency: datetime.timedelta | Period,
        data_frequency_use_window_end: bool,
        data_bundle_source: DataBundleSource,
        asset_service: AssetService,
        bundle_storage_path: str = str(Path(Path.home(), ".ziplime", "data")),
):
    bundle_registry = FileSystemBundleRegistry(base_data_path=bundle_storage_path)
    bundle_service = BundleService(bundle_registry=bundle_registry)
    bundle_storage = FileSystemParquetBundleStorage(base_data_path=bundle_storage_path, compression_level=5)
    bundle_version = str(int(datetime.datetime.now(tz=trading_calendar.tz).timestamp()))

    await bundle_service.ingest_custom_data_bundle(
        date_start=start_date.replace(tzinfo=trading_calendar.tz),
        date_end=end_date.replace(tzinfo=trading_calendar.tz),
        bundle_storage=bundle_storage,
        data_bundle_source=data_bundle_source,
        frequency=frequency,
        data_frequency_use_window_end=data_frequency_use_window_end,
        symbols=symbols,
        name=bundle_name,
        bundle_version=bundle_version,
        trading_calendar=trading_calendar,
        asset_service=asset_service,
    )


async def ingest_market_data(start_date: datetime.datetime, end_date: datetime.datetime,
                             symbols: list[str],
                             trading_calendar: str,
                             bundle_name: str,
                             data_bundle_source: DataBundleSource,
                             asset_service: AssetService,
                             data_frequency: datetime.timedelta = datetime.timedelta(minutes=1)):
    calendar = get_calendar(trading_calendar)
    await _ingest_market_data(start_date=start_date, end_date=end_date, bundle_name=bundle_name,
                                frequency=data_frequency,
                                trading_calendar=calendar, symbols=symbols, data_bundle_source=data_bundle_source,
                                asset_service=asset_service)


async def ingest_custom_data(start_date: datetime.datetime, end_date: datetime.datetime,
                             symbols: list[str],
                             trading_calendar: str,
                             bundle_name: str,
                             data_bundle_source: DataBundleSource,
                             asset_service: AssetService,
                             data_frequency: datetime.timedelta | Period = datetime.timedelta(minutes=1),
                             data_frequency_use_window_end: bool = True):
    calendar = get_calendar(trading_calendar)
    await _ingest_custom_data(start_date=start_date, end_date=end_date, bundle_name=bundle_name,
                              frequency=data_frequency,
                              data_frequency_use_window_end=data_frequency_use_window_end,
                              trading_calendar=calendar, symbols=symbols, data_bundle_source=data_bundle_source,
                              asset_service=asset_service)



from functools import lru_cache

from sqlalchemy.orm import Mapped, relationship

from ziplime.assets.models.asset_model import AssetModel
from ziplime.assets.models.currency_symbol_mapping_model import CurrencySymbolMappingModel


class CurrencyModel(AssetModel):
    __tablename__ = "currencies"
    currency_symbol_mappings: Mapped[list[CurrencySymbolMappingModel]] = relationship("CurrencySymbolMappingModel")

    @lru_cache
    def get_symbol_by_exchange(self, exchange_name: str | None) -> str | None:
        if exchange_name is None:
            symbol_mapping = next(iter(self.currency_symbol_mappings), None)
        else:
            symbol_mapping = next(filter(lambda x: x.exchange == exchange_name, self.currency_symbol_mappings), None)
        if symbol_mapping:
            return symbol_mapping.symbol
        return None



"""
NumericalExpression term.
"""

import re
from itertools import chain
from numbers import Number

import numexpr
from numexpr.necompiler import getExprNames
from numpy import (
    full,
    inf,
)
from ziplime.pipeline.term import Term, ComputableTerm
from ziplime.utils.numpy_utils import bool_dtype

_VARIABLE_NAME_RE = re.compile("^(x_)([0-9]+)$")

# Map from op symbol to equivalent Python magic method name.
ops_to_methods = {
    "+": "__add__",
    "-": "__sub__",
    "*": "__mul__",
    "/": "__div__",
    "%": "__mod__",
    "**": "__pow__",
    "&": "__and__",
    "|": "__or__",
    "^": "__xor__",
    "<": "__lt__",
    "<=": "__le__",
    "==": "__eq__",
    "!=": "__ne__",
    ">=": "__ge__",
    ">": "__gt__",
}
# Map from method name to op symbol.
methods_to_ops = {v: k for k, v in ops_to_methods.items()}

# Map from op symbol to equivalent Python magic method name after flipping
# arguments.
ops_to_commuted_methods = {
    "+": "__radd__",
    "-": "__rsub__",
    "*": "__rmul__",
    "/": "__rdiv__",
    "%": "__rmod__",
    "**": "__rpow__",
    "&": "__rand__",
    "|": "__ror__",
    "^": "__rxor__",
    "<": "__gt__",
    "<=": "__ge__",
    "==": "__eq__",
    "!=": "__ne__",
    ">=": "__le__",
    ">": "__lt__",
}
unary_ops_to_methods = {
    "-": "__neg__",
    "~": "__invert__",
}

UNARY_OPS = {"-"}
MATH_BINOPS = {"+", "-", "*", "/", "**", "%"}
FILTER_BINOPS = {"&", "|"}  # NumExpr doesn't support xor.
COMPARISONS = {"<", "<=", "!=", ">=", ">", "=="}

NUMEXPR_MATH_FUNCS = {
    "sin",
    "cos",
    "tan",
    "arcsin",
    "arccos",
    "arctan",
    "sinh",
    "cosh",
    "tanh",
    "arcsinh",
    "arccosh",
    "arctanh",
    "log",
    "log10",
    "log1p",
    "exp",
    "expm1",
    "sqrt",
    "abs",
}

NPY_MAXARGS = 32


def _ensure_element(tup, elem):
    """
    Create a tuple containing all elements of tup, plus elem.

    Returns the new tuple and the index of elem in the new tuple.
    """
    try:
        return tup, tup.index(elem)
    except ValueError:
        return tuple(chain(tup, (elem,))), len(tup)


class BadBinaryOperator(TypeError):
    """
    Called when a bad binary operation is encountered.

    Parameters
    ----------
    op : str
        The attempted operation
    left : ziplime.computable.Term
        The left hand side of the operation.
    right : ziplime.computable.Term
        The right hand side of the operation.
    """

    def __init__(self, op, left, right):
        super(BadBinaryOperator, self).__init__(
            "Can't compute {left} {op} {right}".format(
                op=op,
                left=type(left).__name__,
                right=type(right).__name__,
            )
        )


def method_name_for_op(op, commute=False):
    """
    Get the name of the Python magic method corresponding to `op`.

    Parameters
    ----------
    op : str {'+','-','*', '/','**','&','|','^','<','<=','==','!=','>=','>'}
        The requested operation.
    commute : bool
        Whether to return the name of an equivalent method after flipping args.

    Returns
    -------
    method_name : str
        The name of the Python magic method corresponding to `op`.
        If `commute` is True, returns the name of a method equivalent to `op`
        with inputs flipped.

    Examples
    --------
    >>> method_name_for_op('+')
    '__add__'
    >>> method_name_for_op('+', commute=True)
    '__radd__'
    >>> method_name_for_op('>')
    '__gt__'
    >>> method_name_for_op('>', commute=True)
    '__lt__'
    """
    if commute:
        return ops_to_commuted_methods[op]
    return ops_to_methods[op]


def unary_op_name(op):
    return unary_ops_to_methods[op]


def is_comparison(op):
    return op in COMPARISONS


class NumericalExpression(ComputableTerm):
    """
    Term binding to a numexpr expression.

    Parameters
    ----------
    expr : string
        A string suitable for passing to numexpr.  All variables in 'expr'
        should be of the form "x_i", where i is the index of the corresponding
        factor input in 'binds'.
    binds : tuple
        A tuple of factors to use as inputs.
    dtype : np.dtype
        The dtype for the expression.
    """

    window_length = 0

    def __new__(cls, expr, binds, dtype):
        # We always allow filters to be used in windowed computations.
        # Otherwise, an expression is window_safe if all its constituents are
        # window_safe.
        window_safe = (dtype == bool_dtype) or all(t.window_safe for t in binds)

        return super(NumericalExpression, cls).__new__(
            cls,
            inputs=binds,
            expr=expr,
            dtype=dtype,
            window_safe=window_safe,
        )

    def _init(self, expr, *args, **kwargs):
        self._expr = expr
        return super(NumericalExpression, self)._init(*args, **kwargs)

    @classmethod
    def _static_identity(cls, expr, *args, **kwargs):
        return (
            super(NumericalExpression, cls)._static_identity(*args, **kwargs),
            expr,
        )

    def _validate(self):
        """
        Ensure that our expression string has variables of the form x_0, x_1,
        ... x_(N - 1), where N is the length of our inputs.
        """
        variable_names, _unused = getExprNames(self._expr, {})
        expr_indices = []
        for name in variable_names:
            if name == "inf":
                continue
            match = _VARIABLE_NAME_RE.match(name)
            if not match:
                raise ValueError("%r is not a valid variable name" % name)
            expr_indices.append(int(match.group(2)))

        expr_indices.sort()
        expected_indices = list(range(len(self.inputs)))
        if expr_indices != expected_indices:
            raise ValueError(
                "Expected %s for variable indices, but got %s"
                % (
                    expected_indices,
                    expr_indices,
                )
            )
        super(NumericalExpression, self)._validate()

    def _compute(self, arrays, dates, assets, mask):
        """
        Compute our stored expression string with numexpr.
        """
        out = full(mask.shape, self.missing_value, dtype=self.dtype)
        # This writes directly into our output buffer.
        numexpr.evaluate(
            self._expr,
            local_dict={"x_%d" % idx: array for idx, array in enumerate(arrays)},
            global_dict={"inf": inf},
            out=out,
        )
        return out

    def _rebind_variables(self, new_inputs):
        """
        Return self._expr with all variables rebound to the indices implied by
        new_inputs.
        """
        expr = self._expr

        # If we have 11+ variables, some of our variable names may be
        # substrings of other variable names. For example, we might have x_1,
        # x_10, and x_100. By enumerating in reverse order, we ensure that
        # every variable name which is a substring of another variable name is
        # processed after the variable of which it is a substring. This
        # guarantees that the substitution of any given variable index only
        # ever affects exactly its own index. For example, if we have variables
        # with indices going up to 100, we will process all of the x_1xx names
        # before x_1x, which will be before x_1, so the substitution of x_1
        # will not affect x_1x, which will not affect x_1xx.
        for idx, input_ in reversed(list(enumerate(self.inputs))):
            old_varname = "x_%d" % idx
            # Temporarily rebind to x_temp_N so that we don't overwrite the
            # same value multiple times.
            temp_new_varname = "x_temp_%d" % new_inputs.index(input_)
            expr = expr.replace(old_varname, temp_new_varname)
        # Clear out the temp variables now that we've finished iteration.
        return expr.replace("_temp_", "_")

    def _merge_expressions(self, other):
        """
        Merge the inputs of two NumericalExpressions into a single input tuple,
        rewriting their respective string expressions to make input names
        resolve correctly.

        Returns a tuple of (new_self_expr, new_other_expr, new_inputs)
        """
        new_inputs = tuple(set(self.inputs).union(other.inputs))
        new_self_expr = self._rebind_variables(new_inputs)
        new_other_expr = other._rebind_variables(new_inputs)
        return new_self_expr, new_other_expr, new_inputs

    def build_binary_op(self, op, other):
        """
        Compute new expression strings and a new inputs tuple for combining
        self and other with a binary operator.
        """
        if isinstance(other, NumericalExpression):
            self_expr, other_expr, new_inputs = self._merge_expressions(other)
        elif isinstance(other, Term):
            self_expr = self._expr
            new_inputs, other_idx = _ensure_element(self.inputs, other)
            other_expr = "x_%d" % other_idx
        elif isinstance(other, Number):
            self_expr = self._expr
            other_expr = str(other)
            new_inputs = self.inputs
        else:
            raise BadBinaryOperator(op, self, other)

        # If the merged inputs would be too many for numexpr, then don't merge
        # them:
        if len(new_inputs) >= NPY_MAXARGS:
            self_expr = "x_0"
            other_expr = "x_1"
            new_inputs = self, other

        return self_expr, other_expr, new_inputs

    @property
    def bindings(self):
        return {"x_%d" % i: input_ for i, input_ in enumerate(self.inputs)}

    def __repr__(self):
        return "{typename}(expr='{expr}', bindings={bindings})".format(
            typename=type(self).__name__,
            expr=self._expr,
            bindings=self.bindings,
        )

    def graph_repr(self):
        """Short repr to use when rendering Pipeline graphs."""

        # Replace any floating point numbers in the expression
        # with their scientific notation
        final = re.sub(
            r"[-+]?\d*\.\d+", lambda x: format(float(x.group(0)), ".2E"), self._expr
        )
        # Graphviz interprets `\l` as "divide label into lines, left-justified"
        return "Expression:\\l  {}\\l".format(
            final,
        )



import enum

from ziplime.domain.column_specification import ColumnSpecification


class FundamentalDataValueType(enum.Enum):
    VALUE = "value"
    TTM = "ttm"
    ADD_VALUE = "add_value"


class FundamentalData(enum.Enum):
    TOTAL_SHARE_HOLDER_EQUITY = "total_share_holder_equity"
    # TOTAL_LIABILITIES = "total_liabilities"
    # TOTAL_ASSETS = "total_assets"
    # SHARES_OUTSTANDING = "shares_outstanding"
    # OPERATING_INCOME = "operating_income"

    ROI = "roi"
    ROE = "roe"
    ROA = "roa"
    REVENUE = "revenue"
    RETURN_ON_TANGIBLE_EQUITY = "return_on_tangible_equity"
    QUICK_RATIO = "quick_ratio"
    PRICE_SALES = "price_sales"
    PRICE_FCF = "price_fcf"
    PE_RATIO = "pe_ratio"
    PRICE_BOOK = "price_book"
    OPERATING_MARGIN = "operating_margin"
    NET_WORTH = "net_worth"
    NET_INCOME = "net_income"
    NUMBER_OF_EMPLOYEES = "number_of_employees"
    GROSS_PROFIT = "gross_profit"
    EPS_FORECAST = "eps_forecast"
    NET_PROFIT_MARGIN = "net_profit_margin"
    DIVIDEND_YIELD = "dividend_yield"
    GROSS_MARGIN = "gross_margin"
    EBITDA = "ebitda"
    EPS = "eps"
    DEBT_EQUITY_RATIO = "debt_equity_ratio"
    CURRENT_RATIO = "current_ratio"
    LONG_TERM_DEBT = "long_term_debt"

    def property(self, value_type: FundamentalDataValueType):
        return self.value + "_" + value_type.value


FUNDAMENTAL_DATA_COLUMNS = [
    ColumnSpecification(name=f"{fd.value}_{fdt.value}", write_type="uint32", original_type='float', scale_factor=1000,
                        scaled_type="int")
    for fd in FundamentalData for fdt in FundamentalDataValueType
]

FUNDAMENTAL_DATA_COLUMN_NAMES = [
    c.name for c in
    FUNDAMENTAL_DATA_COLUMNS
]



from dataclasses import dataclass


@dataclass
class ColumnSpecification:
    """Describes how historical data value is stored"""
    name: str
    original_type: str
    write_type: str
    # used only for floats/Decimals
    scaled_type: str
    scale_factor: int
